import random
import string
from collections import defaultdict

import ehrax as rx
import equinox as eqx
import numpy as np
import pandas as pd
import pytest
from ehrax.testing.common_setup import DATASET_SCHEME_MANAGER


@pytest.fixture(scope="module")
def indexed_dataset(large_dataset: rx.Dataset) -> rx.Dataset:
    return large_dataset._execute_pipeline([rx.SetIndex()], DATASET_SCHEME_MANAGER)


@pytest.fixture(scope="module")
def sample_subject_id(indexed_dataset: rx.Dataset) -> str:
    return random.choice(indexed_dataset.tables.static.index)


@pytest.fixture(scope="module")
def sample_admission_id(indexed_dataset: rx.Dataset) -> str:
    # Get an admission id that exists in all columns.
    candidates = set(indexed_dataset.tables.admissions.index)
    for _, table in indexed_dataset.tables.tables_dict.items():
        if str(rx.COLUMN.admission_id) in table.columns:
            candidates &= set(table[str(rx.COLUMN.admission_id)].values)
    return random.choice(list(candidates))


class TestDatasetTransformation:
    @pytest.fixture(scope="class")
    def removed_subject_admissions_dataset(self, indexed_dataset: rx.Dataset, sample_subject_id: str):
        admissions = indexed_dataset.tables.admissions
        admissions = admissions[admissions[str(rx.COLUMN.subject_id)] != sample_subject_id]
        return eqx.tree_at(lambda x: x.tables.admissions, indexed_dataset, admissions)

    @pytest.fixture(scope="class")
    def removed_no_admission_subjects_RESULTS(self, removed_subject_admissions_dataset: rx.Dataset):
        filtered_dataset, report = rx.DatasetTransformation.filter_no_admission_subjects(
            removed_subject_admissions_dataset, rx.Report()
        )
        return filtered_dataset, report

    @pytest.fixture(scope="class")
    def removed_no_admission_subjects(self, removed_no_admission_subjects_RESULTS) -> rx.Dataset:
        return removed_no_admission_subjects_RESULTS[0]

    @pytest.fixture(scope="class")
    def removed_no_admission_subjects_REPORT(self, removed_no_admission_subjects_RESULTS) -> rx.Report:
        return removed_no_admission_subjects_RESULTS[1]

    @pytest.fixture(scope="class")
    def removed_subject_dataset_unsync(self, indexed_dataset: rx.Dataset, sample_subject_id: str):
        static = indexed_dataset.tables.static
        static = static.drop(index=sample_subject_id)
        return eqx.tree_at(lambda x: x.tables.static, indexed_dataset, static)

    @pytest.fixture(scope="class")
    def removed_admission_dataset_unsync(self, indexed_dataset: rx.Dataset, sample_admission_id: str):
        admissions = indexed_dataset.tables.admissions
        admissions = admissions.drop(index=sample_admission_id)
        return eqx.tree_at(lambda x: x.tables.admissions, indexed_dataset, admissions)

    @pytest.fixture(scope="class")
    def removed_subject_dataset_sync_RESULTS(self, removed_subject_dataset_unsync):
        return rx.DatasetTransformation.synchronize_subjects(removed_subject_dataset_unsync, rx.Report())

    @pytest.fixture(scope="class")
    def removed_subject_dataset_sync(self, removed_subject_dataset_sync_RESULTS) -> rx.Dataset:
        return removed_subject_dataset_sync_RESULTS[0]

    @pytest.fixture(scope="class")
    def removed_subject_dataset_sync_REPORT(self, removed_subject_dataset_sync_RESULTS) -> rx.Report:
        return removed_subject_dataset_sync_RESULTS[1]

    @pytest.fixture(scope="class")
    def removed_admission_dataset_sync_RESUTLS(self, removed_admission_dataset_unsync):
        return rx.DatasetTransformation.synchronize_index(
            removed_admission_dataset_unsync, "admissions", str(rx.COLUMN.admission_id), rx.Report()
        )

    @pytest.fixture(scope="class")
    def removed_admission_dataset_sync(self, removed_admission_dataset_sync_RESUTLS) -> rx.Dataset:
        return removed_admission_dataset_sync_RESUTLS[0]

    @pytest.fixture(scope="class")
    def removed_admission_dataset_sync_REPORT(self, removed_admission_dataset_sync_RESUTLS) -> rx.Report:
        return removed_admission_dataset_sync_RESUTLS[1]

    def test_filter_no_admissions_subjects(self, removed_no_admission_subjects: rx.Dataset, sample_subject_id: str):
        assert sample_subject_id not in removed_no_admission_subjects.tables.static

    def test_synchronize_index_subjects(
        self,
        indexed_dataset: rx.Dataset,
        removed_subject_dataset_unsync: rx.Dataset,
        removed_subject_dataset_sync: rx.Dataset,
        sample_subject_id: str,
    ):
        assert sample_subject_id in indexed_dataset.tables.admissions[str(rx.COLUMN.subject_id)].values
        assert sample_subject_id in removed_subject_dataset_unsync.tables.admissions[str(rx.COLUMN.subject_id)].values
        assert sample_subject_id not in removed_subject_dataset_sync.tables.admissions[str(rx.COLUMN.subject_id)].values
        assert set(removed_subject_dataset_sync.tables.static.index) == set(
            removed_subject_dataset_sync.tables.admissions[str(rx.COLUMN.subject_id)]
        )

    def test_synchronize_index_admissions(
        self,
        removed_admission_dataset_unsync: rx.Dataset,
        removed_admission_dataset_sync: rx.Dataset,
        sample_admission_id: str,
    ):
        for table_name, table in removed_admission_dataset_unsync.tables.tables_dict.items():
            if str(rx.COLUMN.admission_id) in table.columns:
                assert sample_admission_id in table[str(rx.COLUMN.admission_id)].values
                synced_table = getattr(removed_admission_dataset_sync.tables, table_name)
                assert sample_admission_id not in synced_table[str(rx.COLUMN.admission_id)].values
                assert set(synced_table[str(rx.COLUMN.admission_id)]).issubset(
                    set(removed_admission_dataset_sync.tables.admissions.index)
                )

    def test_generated_report1(self, removed_no_admission_subjects_REPORT: rx.Report):
        assert isinstance(removed_no_admission_subjects_REPORT, rx.Report)
        assert len(removed_no_admission_subjects_REPORT) > 0
        assert isinstance(removed_no_admission_subjects_REPORT[0], rx.ReportAttributes)

    def test_serializable_report1(self, removed_no_admission_subjects_REPORT: rx.Report):
        assert all(isinstance(v.as_dict(), dict) for v in removed_no_admission_subjects_REPORT)
        assert all([rx.AbstractConfig.from_dict(v.to_dict()).equals(v) for v in removed_no_admission_subjects_REPORT])


class TestCastTimestamps:
    @pytest.fixture(scope="class")
    def str_timestamps_dataset(self, indexed_dataset: rx.Dataset):
        for table_name, time_cols in indexed_dataset.config.columns.time_cols.items():
            table = indexed_dataset.tables.tables_dict[table_name].copy()
            for col in time_cols:
                table[col] = table[col].astype(str)
            indexed_dataset = eqx.tree_at(lambda x: getattr(x.tables, table_name), indexed_dataset, table)
        return indexed_dataset

    @pytest.fixture(scope="class")
    def casted_timestamps_dataset(self, str_timestamps_dataset: rx.Dataset):
        return rx.CastTimestamps.apply(str_timestamps_dataset, DATASET_SCHEME_MANAGER, rx.Report())[0]

    def test_cast_timestamps(self, str_timestamps_dataset: rx.Dataset, casted_timestamps_dataset: rx.Dataset):
        for table_name, time_cols in str_timestamps_dataset.config.columns.time_cols.items():
            table1 = str_timestamps_dataset.tables.tables_dict[table_name]
            table2 = casted_timestamps_dataset.tables.tables_dict[table_name]
            for col in time_cols:
                assert table1[col].dtype == np.dtype("O")
                assert table2[col].dtype == np.dtype("datetime64[ns]")


class TestFilterUnsupportedCodes:
    @pytest.fixture(scope="class")
    def dataset_with_unsupported_codes(self, indexed_dataset: rx.Dataset) -> tuple[rx.Dataset, dict[str, set[str]]]:
        unsupported_codes = {}
        for table_name, code_col in indexed_dataset.config.columns.code_column.items():
            table = indexed_dataset.tables.tables_dict[table_name]
            unsupported_code = f"UNSUPPORTED_CODE_{''.join(random.choices(string.ascii_uppercase, k=5))}"
            table.loc[table.index[0], code_col] = unsupported_code
            unsupported_codes[table_name] = unsupported_code
            indexed_dataset = eqx.tree_at(lambda x: getattr(x.tables, table_name), indexed_dataset, table)
        return indexed_dataset, unsupported_codes

    @pytest.fixture(scope="class")
    def filtered_dataset(self, dataset_with_unsupported_codes: tuple[rx.Dataset, dict[str, set[str]]]) -> rx.Dataset:
        dataset, _ = dataset_with_unsupported_codes
        return rx.FilterUnsupportedCodes.apply(dataset, DATASET_SCHEME_MANAGER, rx.Report())[0]

    def test_filter_unsupported_codes(
        self, dataset_with_unsupported_codes: tuple[rx.Dataset, dict[str, set[str]]], filtered_dataset: rx.Dataset
    ):
        unfiltered_dataset, unsupported_codes = dataset_with_unsupported_codes
        for table_name, code_col in filtered_dataset.config.columns.code_column.items():
            (code_col,) = code_col
            assert unsupported_codes[table_name] in getattr(unfiltered_dataset.tables, table_name)[code_col].values
            assert unsupported_codes[table_name] not in getattr(filtered_dataset.tables, table_name)[code_col].values


class TestSetRelativeTimes:
    @pytest.fixture(scope="class")
    def relative_times_dataset(self, indexed_dataset: rx.Dataset):
        return rx.SetAdmissionRelativeTimes.apply(indexed_dataset, DATASET_SCHEME_MANAGER, rx.Report())[0]

    @pytest.fixture(scope="class")
    def admission_los_table(self, indexed_dataset: rx.Dataset):
        admissions = indexed_dataset.tables.admissions.copy()
        admissions["los_hours"] = (
            admissions[str(rx.COLUMN.end_time)] - admissions[str(rx.COLUMN.start_time)]
        ).dt.total_seconds() / (60 * 60)
        return admissions[["los_hours"]]

    def test_set_relative_times(
        self, indexed_dataset: rx.Dataset, relative_times_dataset: rx.Dataset, admission_los_table: pd.DataFrame
    ):
        for table_name, time_cols in indexed_dataset.config.columns.time_cols.items():
            if table_name in ("admissions", "static"):
                continue
            table = getattr(relative_times_dataset.tables, table_name)
            table = table.merge(admission_los_table, left_on=str(rx.COLUMN.admission_id), right_index=True)
            for col in time_cols:
                assert table[col].dtype == float
                assert table[col].min() >= 0
                assert all(table[col] <= table["los_hours"])


class TestFilterSubjectsWithNegativeAdmissionInterval:
    @pytest.fixture(scope="class")
    def dataset(self, indexed_dataset: rx.Dataset) -> rx.Dataset:
        return rx.FilterSubjectsNegativeAdmissionLengths.apply(indexed_dataset, DATASET_SCHEME_MANAGER, rx.Report())[0]

    @pytest.fixture(scope="class")
    def dataset_inverted_admission(self, dataset: rx.Dataset, sample_admission_id: str) -> rx.Dataset:
        admissions = dataset.tables.admissions.copy()
        c_admittime = dataset.config.columns.admissions.start_time
        c_dischtime = dataset.config.columns.admissions.end_time
        admittime = admissions.loc[sample_admission_id, c_admittime]
        dischtime = admissions.loc[sample_admission_id, c_dischtime]
        admissions.loc[sample_admission_id, c_admittime] = dischtime
        admissions.loc[sample_admission_id, c_dischtime] = admittime
        return eqx.tree_at(lambda x: x.tables.admissions, dataset, admissions)

    @pytest.fixture(scope="class")
    def filtered_dataset(self, dataset_inverted_admission: rx.Dataset):
        return rx.FilterSubjectsNegativeAdmissionLengths.apply(
            dataset_inverted_admission, DATASET_SCHEME_MANAGER, rx.Report()
        )[0]

    def test_filter_subjects_negative_admission_length(
        self, dataset_inverted_admission: rx.Dataset, filtered_dataset: rx.Dataset, sample_admission_id: str
    ):
        admissions0 = dataset_inverted_admission.tables.admissions
        static0 = dataset_inverted_admission.tables.static
        admissions1 = filtered_dataset.tables.admissions
        static1 = filtered_dataset.tables.static

        assert admissions0.shape[0] > admissions1.shape[0]
        assert static0.shape[0] == static1.shape[0] + 1
        assert (
            admissions0.loc[sample_admission_id, str(rx.COLUMN.start_time)]
            > admissions0.loc[sample_admission_id, str(rx.COLUMN.end_time)]
        )
        assert any(admissions0[str(rx.COLUMN.start_time)] > admissions0[str(rx.COLUMN.end_time)])
        assert all(admissions1[str(rx.COLUMN.start_time)] <= admissions1[str(rx.COLUMN.end_time)])
        assert sample_admission_id in admissions0.index
        assert sample_admission_id not in admissions1.index


class TestOverlappingAdmissions:
    def _generate_admissions_from_pattern(self, pattern: list[str]) -> pd.DataFrame:
        if len(pattern) == 0:
            return pd.DataFrame(columns=[str(rx.COLUMN.start_time), str(rx.COLUMN.end_time)])
        random_monotonic_positive_integers = np.random.randint(1, 30, size=len(pattern)).cumsum()
        sequence_dates = list(
            map(lambda x: pd.Timestamp.today().normalize() + pd.Timedelta(days=x), random_monotonic_positive_integers)
        )
        event_times = dict(zip(pattern, sequence_dates))
        admission_time = {k: v for k, v in event_times.items() if k.startswith("A")}
        discharge_time = {k.replace("D", "A"): v for k, v in event_times.items() if k.startswith("D")}
        admissions = pd.DataFrame(index=list(admission_time.keys()))
        admissions[str(rx.COLUMN.start_time)] = admissions.index.map(admission_time)
        admissions[str(rx.COLUMN.end_time)] = admissions.index.map(discharge_time)
        # shuffle the rows shuffled.
        return admissions.sample(frac=1)

    @pytest.fixture(
        params=[
            # Each line below has an ordered list of events labeled Ak and Dk, where Ak is the k-th admission event
            # and Dk is the discharge of k-th admission; and a dictionary of the
            # expected {super_interval: [merged_subintervals]}.
            # In the database, it is possible to have overlapping admissions, so we merge them.
            # A1 and D1 represent the admission and discharge time of a particular admission record in the database.
            # A2 and D2 mean the same, and that A2 happened after A1 (and not necessarily after D1).
            (["A1", "A2", "A3", "D1", "D3", "D2"], {"A1": ["A2", "A3"]}),
            (["A1", "A2", "A3", "D1", "D2", "D3"], {"A1": ["A2", "A3"]}),
            (["A1", "A2", "A3", "D2", "D1", "D3"], {"A1": ["A2", "A3"]}),
            (["A1", "A2", "A3", "D2", "D3", "D1"], {"A1": ["A2", "A3"]}),
            (["A1", "A2", "A3", "D3", "D1", "D2"], {"A1": ["A2", "A3"]}),
            (["A1", "A2", "A3", "D3", "D2", "D1"], {"A1": ["A2", "A3"]}),
            ##
            (["A1", "A2", "D1", "A3", "D3", "D2"], {"A1": ["A2", "A3"]}),
            (["A1", "A2", "D1", "A3", "D2", "D3"], {"A1": ["A2", "A3"]}),
            ##
            (["A1", "A2", "D1", "D2", "A3", "D3"], {"A1": ["A2"]}),
            (["A1", "A2", "D1", "D2"], {"A1": ["A2"]}),
            ##
            (["A1", "A2", "D2", "A3", "D1", "D3"], {"A1": ["A2", "A3"]}),
            (["A1", "A2", "D2", "A3", "D3", "D1"], {"A1": ["A2", "A3"]}),
            (["A1", "A2", "D2", "D1", "A3", "D3"], {"A1": ["A2"]}),
            ##
            (["A1", "D1", "A2", "A3", "D2", "D3"], {"A2": ["A3"]}),
            (["A1", "D1", "A2", "A3", "D3", "D2"], {"A2": ["A3"]}),
            (["A1", "D1", "A2", "D2", "A3", "D3"], {}),
            ##
            (["A1", "D1"], {}),
            ([], {}),
            (["A1", "D1", "A2", "A3", "D3", "A4", "D2", "D4"], {"A2": ["A3", "A4"]}),
            (["A1", "A2", "D2", "D1", "A3", "A4", "D3", "D4"], {"A1": ["A2"], "A3": ["A4"]}),
        ],
        scope="class",
    )
    def admission_pattern_with_expected_out(self, request):
        return request.param

    @pytest.fixture(scope="class")
    def admission_pattern(self, admission_pattern_with_expected_out):
        return admission_pattern_with_expected_out[0]

    @pytest.fixture(scope="class")
    def expected_out(self, admission_pattern_with_expected_out):
        return admission_pattern_with_expected_out[1]

    @pytest.fixture(scope="class")
    def admissions_table(self, admission_pattern) -> pd.DataFrame:
        return self._generate_admissions_from_pattern(admission_pattern)

    @pytest.fixture(scope="class")
    def superset_admissions_dictionary(self, admissions_table: pd.DataFrame) -> dict[str, list[str]]:
        sub2sup = rx.MergeOverlappingAdmissions._collect_overlaps(admissions_table)
        sup2sub = defaultdict(list)
        for sub, sup in sub2sup.items():
            sup2sub[sup].append(sub)
        return sup2sub

    def test_overlapping_cases(self, superset_admissions_dictionary, expected_out):
        assert superset_admissions_dictionary == expected_out

    @pytest.fixture(scope="class")
    def sample_admission_ids_map(self, indexed_dataset: rx.Dataset):
        index = indexed_dataset.tables.admissions.index
        return {
            index[1]: index[0],
            index[2]: index[0],
            index[3]: index[0],
            index[5]: index[4],
            index[6]: index[4],
        }

    @pytest.fixture(scope="class")
    def merged_admissions_dataset(self, indexed_dataset: rx.Dataset, sample_admission_ids_map: dict[str, str]):
        return rx.MergeOverlappingAdmissions._admissions_map_admission_ids(
            indexed_dataset, sample_admission_ids_map, rx.Report()
        )[0]

    @pytest.fixture(scope="class")
    def mapped_tables_dataset(self, indexed_dataset: rx.Dataset, sample_admission_ids_map: dict[str, str]):
        return rx.MergeOverlappingAdmissions._tables_map_admission_ids(
            indexed_dataset, sample_admission_ids_map, rx.Report()
        )[0]

    def test_merge_admission_ids(
        self,
        indexed_dataset: rx.Dataset,
        merged_admissions_dataset: rx.Dataset,
        sample_admission_ids_map: dict[str, str],
    ):
        admissions0 = indexed_dataset.tables.admissions
        admissions1 = merged_admissions_dataset.tables.admissions

        assert len(admissions0) == len(admissions1) + len(sample_admission_ids_map)
        assert set(admissions1.index).issubset(set(admissions0.index))

    def test_map_admission_ids(
        self, indexed_dataset: rx.Dataset, mapped_tables_dataset: rx.Dataset, sample_admission_ids_map: dict[str, str]
    ):
        admissions0 = indexed_dataset.tables.admissions
        admissions1 = mapped_tables_dataset.tables.admissions
        for table_name, table1 in mapped_tables_dataset.tables.tables_dict.items():
            table0 = getattr(indexed_dataset.tables, table_name)
            if str(rx.COLUMN.admission_id) in table1.columns:
                assert len(table1) == len(table0)
                assert set(table1[str(rx.COLUMN.admission_id)]) - set(admissions1.index.values) == set()
                assert set(table0[str(rx.COLUMN.admission_id)]) - set(admissions0.index.values) == set()
                assert set(table1[str(rx.COLUMN.admission_id)]) - set(table0[str(rx.COLUMN.admission_id)]) == set()

    @pytest.fixture(scope="class")
    def large_dataset_overlaps_dictionary(self, indexed_dataset: rx.Dataset):
        admissions = indexed_dataset.tables.admissions

        sub2sup = {
            adm_id: super_adm_id
            for _, subject_adms in admissions.groupby(str(rx.COLUMN.subject_id))
            for adm_id, super_adm_id in rx.MergeOverlappingAdmissions._collect_overlaps(subject_adms).items()
        }

        if len(sub2sup) == 0:
            assert 0, "No overlapping admissions in rx.Dataset."

        return sub2sup

    @pytest.fixture(scope="class")
    def merged_overlapping_admission_dataset(self, indexed_dataset: rx.Dataset):
        return rx.MergeOverlappingAdmissions.apply(indexed_dataset, DATASET_SCHEME_MANAGER, rx.Report())[0]

    @pytest.fixture(scope="class")
    def removed_overlapping_admission_subjects_dataset(self, indexed_dataset: rx.Dataset):
        return rx.RemoveSubjectsWithOverlappingAdmissions.apply(indexed_dataset, DATASET_SCHEME_MANAGER, rx.Report())[0]

    def test_process_overlapping_admissions(
        self,
        indexed_dataset: rx.Dataset,
        large_dataset_overlaps_dictionary: dict[str, str],
        merged_overlapping_admission_dataset: rx.Dataset,
        removed_overlapping_admission_subjects_dataset: rx.Dataset,
    ):
        admissions0 = indexed_dataset.tables.admissions
        admissions_m = merged_overlapping_admission_dataset.tables.admissions
        admissions_r = removed_overlapping_admission_subjects_dataset.tables.admissions

        assert len(admissions0) == len(admissions_m) + len(large_dataset_overlaps_dictionary)
        assert len(admissions_m) > len(admissions_r)
        assert len(merged_overlapping_admission_dataset.tables.static) > len(
            removed_overlapping_admission_subjects_dataset.tables.static
        )

        for table_name, table0 in indexed_dataset.tables.tables_dict.items():
            if table_name in ("static", "admissions"):
                continue
            table_m = getattr(merged_overlapping_admission_dataset.tables, table_name)
            table_r = getattr(removed_overlapping_admission_subjects_dataset.tables, table_name)
            assert len(table_m) == len(table0)
            assert set(table_m[str(rx.COLUMN.admission_id)]) - set(admissions_m.index.values) == set()
            assert set(table0[str(rx.COLUMN.admission_id)]) - set(admissions0.index.values) == set()
            assert set(table_m[str(rx.COLUMN.admission_id)]) - set(table0[str(rx.COLUMN.admission_id)]) == set()
            assert len(table_r) <= len(table0)
            assert (
                set(table_r[str(rx.COLUMN.admission_id)]).intersection(
                    set(large_dataset_overlaps_dictionary.keys()) | set(large_dataset_overlaps_dictionary.values())
                )
                == set()
            )


def test_select_subjects_with_observation(indexed_dataset: rx.Dataset):
    # assert False
    pass


class TestClampTimestamps:
    @pytest.fixture(scope="class")
    def shifted_timestamps_dataset(self, indexed_dataset: rx.Dataset):
        if any(len(getattr(indexed_dataset.tables, k)) == 0 for k in indexed_dataset.config.columns.time_cols.keys()):
            raise pytest.skip("No temporal data in rx.Dataset.")

        admissions = indexed_dataset.tables.admissions

        c_admission_id = indexed_dataset.config.columns.admissions.admission_id
        c_admittime = indexed_dataset.config.columns.admissions.start_time
        c_dischtime = indexed_dataset.config.columns.admissions.end_time
        admission_id = admissions.index[0]
        admittime = admissions.loc[admission_id, c_admittime]
        dischtime = admissions.loc[admission_id, c_dischtime]
        if "obs" in indexed_dataset.tables.tables_dict:
            assert indexed_dataset.config.columns.obs is not None
            c_time = indexed_dataset.config.columns.obs.time

            obs = indexed_dataset.tables.obs.copy()
            admission_obs = obs[obs[c_admission_id] == admission_id]
            if len(admission_obs) > 0:
                obs.loc[admission_obs.index[0], c_time] = dischtime + pd.Timedelta(days=1)
            if len(admission_obs) > 1:
                obs.loc[admission_obs.index[1], c_time] = admittime + pd.Timedelta(days=-1)
            indexed_dataset = eqx.tree_at(lambda x: x.tables.obs, indexed_dataset, obs)
        for k in ("hosp_procedures", "icu_procedures", "icu_inputs"):
            if k in indexed_dataset.tables.tables_dict:
                c_starttime = getattr(indexed_dataset.config.columns, k).start_time
                c_endtime = getattr(indexed_dataset.config.columns, k).end_time
                procedures = getattr(indexed_dataset.tables, k).copy()

                admission_procedures = procedures[procedures[c_admission_id] == admission_id]
                if len(admission_procedures) > 0:
                    procedures.loc[admission_procedures.index[0], c_starttime] = admittime + pd.Timedelta(days=-1)
                    procedures.loc[admission_procedures.index[0], c_endtime] = dischtime + pd.Timedelta(days=1)

                if len(admission_procedures) > 1:
                    procedures.loc[admission_procedures.index[1], c_starttime] = dischtime + pd.Timedelta(days=1)
                    procedures.loc[admission_procedures.index[1], c_endtime] = dischtime + pd.Timedelta(days=2)

                if len(admission_procedures) > 2:
                    procedures.loc[admission_procedures.index[2], c_starttime] = admittime + pd.Timedelta(days=-2)
                    procedures.loc[admission_procedures.index[2], c_endtime] = admittime + pd.Timedelta(days=-1)

                indexed_dataset = eqx.tree_at(lambda x: getattr(x.tables, k), indexed_dataset, procedures)

        return indexed_dataset

    @pytest.fixture(scope="class")
    def fixed_dataset(self, shifted_timestamps_dataset: rx.Dataset):
        return rx.FilterClampTimestampsToAdmissionInterval.apply(
            shifted_timestamps_dataset, DATASET_SCHEME_MANAGER, rx.Report()
        )[0]

    def test_clamp_timestamps_to_admission_interval(
        self, shifted_timestamps_dataset: rx.Dataset, fixed_dataset: rx.Dataset
    ):
        admissions = shifted_timestamps_dataset.tables.admissions

        admission_id = admissions.index[0]
        admittime = admissions.loc[admission_id, str(rx.COLUMN.start_time)]
        dischtime = admissions.loc[admission_id, str(rx.COLUMN.end_time)]

        if "obs" in shifted_timestamps_dataset.tables.tables_dict:
            assert shifted_timestamps_dataset.config.columns.obs is not None

            c_time = shifted_timestamps_dataset.config.columns.obs.time

            obs0 = shifted_timestamps_dataset.tables.obs
            obs1 = fixed_dataset.tables.obs
            assert obs0 is not None
            assert obs1 is not None

            admission_obs0 = obs0[obs0[str(rx.COLUMN.admission_id)] == admission_id]
            admission_obs1 = obs1[obs1[str(rx.COLUMN.admission_id)] == admission_id]
            if len(admission_obs0) > 0:
                assert len(admission_obs0) > len(admission_obs1)
                assert not admission_obs0[c_time].between(admittime, dischtime).all()
                assert admission_obs1[c_time].between(admittime, dischtime).all()

        for k in ("hosp_procedures", "icu_procedures", "icu_inputs"):
            if k in shifted_timestamps_dataset.tables.tables_dict:
                c_starttime = getattr(shifted_timestamps_dataset.config.columns, k).start_time
                c_endtime = getattr(shifted_timestamps_dataset.config.columns, k).end_time
                procedures0 = getattr(shifted_timestamps_dataset.tables, k)
                procedures1 = getattr(fixed_dataset.tables, k)

                admission_procedures0 = procedures0[procedures0[str(rx.COLUMN.admission_id)] == admission_id]
                admission_procedures1 = procedures1[procedures1[str(rx.COLUMN.admission_id)] == admission_id]
                if len(admission_procedures0) > 0:
                    assert len(admission_procedures0) >= len(admission_procedures1)
                    assert admission_procedures1[c_starttime].between(admittime, dischtime).all()
                    assert admission_procedures1[c_endtime].between(admittime, dischtime).all()

                if len(admission_procedures0) > 1:
                    assert len(admission_procedures0) > len(admission_procedures1)


class TestFilterAdmissionsWithNoDiagnoses:
    ADMISSION_ID = list(map(str, range(10)))
    ADMISSION_ID_WITH_NO_DIAGNOSIS = ["2", "5"]

    @pytest.fixture(
        scope="class", params=[(ADMISSION_ID, ADMISSION_ID_WITH_NO_DIAGNOSIS), (ADMISSION_ID, []), ([], [])]
    )
    def admission_id_admission_id_with_no_diagnosis(self, request) -> tuple[list[str], list[str]]:
        return request.param

    @pytest.fixture(scope="class")
    def admission_id(self, admission_id_admission_id_with_no_diagnosis: tuple[list[str], list[str]]) -> list[str]:
        return admission_id_admission_id_with_no_diagnosis[0]

    @pytest.fixture(scope="class")
    def admission_id_with_no_diagnosis(
        self, admission_id_admission_id_with_no_diagnosis: tuple[list[str], list[str]]
    ) -> list[str]:
        return admission_id_admission_id_with_no_diagnosis[1]

    @pytest.fixture(scope="class")
    def static(self) -> pd.DataFrame:
        return pd.DataFrame({rx.COLUMN.subject_id: list(map(str, range(100, 110)))}).set_index(rx.COLUMN.subject_id)

    @pytest.fixture(scope="class")
    def admissions(self, static: pd.DataFrame, admission_id: list[str]) -> pd.DataFrame:
        return pd.DataFrame(
            {
                rx.COLUMN.subject_id: random.choices(static.index, k=len(admission_id)),
                rx.COLUMN.admission_id: admission_id,
            }
        ).set_index(rx.COLUMN.admission_id)

    @pytest.fixture(scope="class")
    def dx_discharge(
        self, admissions: pd.DataFrame, admission_id: list[str], admission_id_with_no_diagnosis: list[str]
    ) -> pd.DataFrame:
        return pd.DataFrame({rx.COLUMN.admission_id: list(set(admission_id) - set(admission_id_with_no_diagnosis))})

    @pytest.fixture(scope="class")
    def dataset(self, static: pd.DataFrame, dx_discharge: pd.DataFrame, admissions: pd.DataFrame) -> rx.Dataset:
        return rx.Dataset(
            tables=rx.DatasetTables(static=static, admissions=admissions, dx_discharge=dx_discharge),
            config=rx.DatasetConfig(scheme=rx.DatasetSchemeConfig()),
        )

    @pytest.fixture(scope="class")
    def filtered_dataset(self, dataset: rx.Dataset) -> rx.Dataset:
        updated_dataset, _ = rx.FilterAdmissionsWithNoDiagnoses.apply(dataset, None, rx.Report())
        return updated_dataset

    def test_transformation(
        self,
        dataset: rx.Dataset,
        filtered_dataset: rx.Dataset,
        admission_id: list[str],
        admission_id_with_no_diagnosis: list[str],
    ):
        assert set(admission_id).issubset(dataset.tables.admissions.index)
        assert set(filtered_dataset.tables.admissions.index) == set(admission_id) - set(admission_id_with_no_diagnosis)


class TestFilterSubjectsWithSingleOrNoAdmission:
    SUBJECT_ID = list(map(str, range(10)))
    SUBJECT_ID_WITH_NO_ADMISSIONS = ["4", "5"]

    @pytest.fixture(scope="class", params=[(SUBJECT_ID, SUBJECT_ID_WITH_NO_ADMISSIONS), (SUBJECT_ID, []), ([], [])])
    def subject_id_subject_id_with_1_admissions(self, request) -> tuple[list[str], list[str]]:
        return request.param

    @pytest.fixture(scope="class")
    def subject_id(self, subject_id_subject_id_with_1_admissions: tuple[list[str], list[str]]) -> list[str]:
        return subject_id_subject_id_with_1_admissions[0]

    @pytest.fixture(scope="class")
    def subject_id_with_1_admission(
        self, subject_id_subject_id_with_1_admissions: tuple[list[str], list[str]]
    ) -> list[str]:
        return subject_id_subject_id_with_1_admissions[1]

    @pytest.fixture(scope="class")
    def static(self, subject_id: list[str]) -> pd.DataFrame:
        return pd.DataFrame({rx.COLUMN.subject_id: subject_id}).set_index(rx.COLUMN.subject_id)

    @pytest.fixture(scope="class")
    def admissions(
        self, static: pd.DataFrame, subject_id: list[str], subject_id_with_1_admission: list[str]
    ) -> pd.DataFrame:
        admission_id_1 = list(map(str, range(len(subject_id))))
        df_1 = pd.DataFrame({rx.COLUMN.subject_id: subject_id, rx.COLUMN.admission_id: admission_id_1})
        n1 = len(subject_id)
        subject_w_2_admissions = set(subject_id) - set(subject_id_with_1_admission)
        admission_id_2 = list(map(str, range(n1, n1 + len(subject_w_2_admissions))))
        df_2 = pd.DataFrame(
            {rx.COLUMN.subject_id: list(subject_w_2_admissions), rx.COLUMN.admission_id: admission_id_2}
        )

        return pd.concat([df_1, df_2], ignore_index=True).set_index(rx.COLUMN.admission_id)

    @pytest.fixture(scope="class")
    def dataset(self, static: pd.DataFrame, admissions: pd.DataFrame) -> rx.Dataset:
        return rx.Dataset(
            tables=rx.DatasetTables(static=static, admissions=admissions),
            config=rx.DatasetConfig(scheme=rx.DatasetSchemeConfig()),
        )

    @pytest.fixture(scope="class")
    def filtered_dataset(self, dataset: rx.Dataset) -> rx.Dataset:
        updated_dataset, _ = rx.FilterSubjectsWithSingleOrNoAdmission.apply(dataset, None, rx.Report())
        return updated_dataset

    def test_transformation(
        self,
        dataset: rx.Dataset,
        filtered_dataset: rx.Dataset,
        subject_id: list[str],
        subject_id_with_1_admission: list[str],
    ):
        assert set(subject_id).issubset(dataset.tables.static.index)
        assert set(filtered_dataset.tables.static.index) == set(subject_id) - set(subject_id_with_1_admission)


class TestFilterSubjectsWithLongAdmission:
    SUBJECT_ID = list(map(str, range(10)))
    SUBJECT_ID_WITH_LONG_ADMISSIONS = ["4", "5"]
    MAX_LOS_DAYS = 10.0

    @pytest.fixture(scope="class", params=[(SUBJECT_ID, SUBJECT_ID_WITH_LONG_ADMISSIONS), (SUBJECT_ID, []), ([], [])])
    def subject_id_subject_id_with_long_admissions(self, request) -> tuple[list[str], list[str]]:
        return request.param

    @pytest.fixture(scope="class")
    def subject_id(self, subject_id_subject_id_with_long_admissions: tuple[list[str], list[str]]) -> list[str]:
        return subject_id_subject_id_with_long_admissions[0]

    @pytest.fixture(scope="class")
    def subject_id_with_long_admission(
        self, subject_id_subject_id_with_long_admissions: tuple[list[str], list[str]]
    ) -> list[str]:
        return subject_id_subject_id_with_long_admissions[1]

    @pytest.fixture(scope="class")
    def static(self, subject_id: list[str]) -> pd.DataFrame:
        return pd.DataFrame({rx.COLUMN.subject_id: subject_id}).set_index(rx.COLUMN.subject_id)

    @pytest.fixture(scope="class")
    def admissions(
        self, static: pd.DataFrame, subject_id: list[str], subject_id_with_long_admission: list[str]
    ) -> pd.DataFrame:
        admission_id_1 = list(map(str, range(len(subject_id))))
        df_1 = pd.DataFrame(
            {
                rx.COLUMN.subject_id: subject_id,
                rx.COLUMN.admission_id: admission_id_1,
                rx.COLUMN.start_time: pd.Timestamp.now(),
                rx.COLUMN.end_time: pd.Timestamp.now()
                + pd.Timedelta(days=random.uniform(0.0, self.MAX_LOS_DAYS * 0.99)),
            }
        )
        n1 = len(subject_id)
        admission_id_2 = list(map(str, range(n1, n1 + len(subject_id_with_long_admission))))
        df_2 = pd.DataFrame(
            {
                rx.COLUMN.subject_id: list(subject_id_with_long_admission),
                rx.COLUMN.admission_id: admission_id_2,
                rx.COLUMN.start_time: pd.Timestamp.now(),
                rx.COLUMN.end_time: pd.Timestamp.now()
                + pd.Timedelta(days=random.uniform(self.MAX_LOS_DAYS * 1.01, self.MAX_LOS_DAYS * 2.0)),
            }
        )
        return pd.concat([df_1, df_2], ignore_index=True).set_index(rx.COLUMN.admission_id)

    @pytest.fixture(scope="class")
    def dataset_no_config(self, static: pd.DataFrame, admissions: pd.DataFrame) -> rx.Dataset:
        return rx.Dataset(
            tables=rx.DatasetTables(static=static, admissions=admissions),
            config=rx.DatasetConfig(scheme=rx.DatasetSchemeConfig()),
        )

    @pytest.fixture(scope="class")
    def dataset(self, dataset_no_config: rx.Dataset) -> rx.Dataset:
        return eqx.tree_at(
            lambda x: x.config.select_subjects_with_short_admissions,
            dataset_no_config,
            self.MAX_LOS_DAYS,
            is_leaf=lambda x: x is None,
        )

    @pytest.fixture(scope="class")
    def filtered_dataset_no_config(self, dataset_no_config: rx.Dataset) -> rx.Dataset:
        updated, _ = rx.FilterSubjectsWithLongAdmission.apply(dataset_no_config, None, rx.Report())
        return updated

    @pytest.fixture(scope="class")
    def filtered_dataset(self, dataset: rx.Dataset) -> rx.Dataset:
        updated_dataset, _ = rx.FilterSubjectsWithLongAdmission.apply(dataset, None, rx.Report())
        return updated_dataset

    def test_transformation(
        self,
        dataset: rx.Dataset,
        filtered_dataset: rx.Dataset,
        filtered_dataset_no_config: rx.Dataset,
        subject_id: list[str],
        subject_id_with_long_admission: list[str],
    ):
        assert set(subject_id).issubset(dataset.tables.static.index)
        assert set(filtered_dataset.tables.static.index) == set(subject_id) - set(subject_id_with_long_admission)
        assert dataset.tables.equals(filtered_dataset_no_config.tables)


class TestSqueezeToStandardColumns:
    @pytest.fixture(scope="class")
    def static(self) -> pd.DataFrame:
        return pd.DataFrame(
            {
                rx.COLUMN.subject_id: list(range(100)),
                rx.COLUMN.gender: list(reversed(range(100))),
                rx.COLUMN.date_of_birth: ["x"] * 100,
                "other_column": ["y"] * 100,
            }
        )

    @pytest.fixture(scope="class")
    def indexed_static(self, static: pd.DataFrame) -> pd.DataFrame:
        return static.set_index(rx.COLUMN.subject_id)

    @pytest.fixture(scope="class")
    def dataset(self, static: pd.DataFrame) -> rx.Dataset:
        return rx.Dataset(
            tables=rx.DatasetTables(static=static, admissions=pd.DataFrame(columns=["admission_id"])),
            config=rx.DatasetConfig(scheme=rx.DatasetSchemeConfig()),
        )

    @pytest.fixture(scope="class")
    def indexed_dataset(self, indexed_static: pd.DataFrame) -> rx.Dataset:
        return rx.Dataset(
            tables=rx.DatasetTables(static=indexed_static, admissions=pd.DataFrame(columns=["admission_id"])),
            config=rx.DatasetConfig(scheme=rx.DatasetSchemeConfig()),
        )

    @pytest.fixture(scope="class")
    def indexed_transformed(self, indexed_dataset: rx.Dataset) -> rx.Dataset:
        updated, _ = rx.SqueezeToStandardColumns.apply(indexed_dataset, None, rx.Report())
        return updated

    @pytest.fixture(scope="class")
    def transformed(self, dataset: rx.Dataset) -> rx.Dataset:
        updated, _ = rx.SqueezeToStandardColumns.apply(dataset, None, rx.Report())
        return updated

    def test_transformation(
        self, transformed: rx.Dataset, indexed_transformed: rx.Dataset, dataset: rx.Dataset, indexed_dataset: rx.Dataset
    ):
        for ds in (dataset, indexed_dataset):
            assert "other_column" in ds.tables.static.columns
        for ds in (transformed, indexed_transformed):
            assert "other_column" not in ds.tables.static.columns
        assert set(transformed.tables.static.columns) - set(indexed_transformed.tables.static.columns) == {"subject_id"}
        assert set(dataset.tables.static.columns) - set(transformed.tables.static.columns) == {"other_column"}
        assert set(dataset.tables.static.columns) - set(indexed_transformed.tables.static.columns) == {
            "other_column",
            "subject_id",
        }



"""Extract diagnostic/procedure information of CCS files into new
data structures to support conversion between CCS and ICD9."""

import logging
import math
import re
from abc import ABCMeta, abstractmethod
from collections import defaultdict, OrderedDict
from collections.abc import Collection, Iterable, Mapping, Sized
from dataclasses import dataclass
from functools import cached_property
from types import MappingProxyType
from typing import cast, Self

import numpy as np
import pandas as pd

from ._literals import AggregationLiteral, NumericalTypeHint
from .base import AbstractVxData
from .freezer import FrozenDict1N, FrozenDict1NM, FrozenDict11
from .utils import Array, dataframe_log, load_config, resources_path, tqdm_constructor


class CodesVector(AbstractVxData):
    vec: Array
    scheme: str

    def __init__(self, vec: Array, scheme: str):
        self.vec = vec
        self.scheme = scheme

    def __hash__(self) -> int:
        return hash(self.scheme + str(self.vec))

    @classmethod
    def empty_like(cls, other: Self) -> Self:
        """
        Creates an empty CodesVector with the same shape as the given CodesVector.

        Args:
            other (CodesVector): the CodesVector to mimic.

        Returns:
            CodesVector: the empty CodesVector.
        """
        return cls(np.zeros_like(other.vec), other.scheme)

    def union(self, other: Self) -> Self:
        """
        Returns the union of the current CodesVector with another CodesVector.

        Args:
            other (CodesVector): the other CodesVector to union with.

        Returns:
            CodesVector: the union of the two CodesVectors.
        """
        assert self.scheme == other.scheme, "Schemes should be the same."
        assert self.vec.dtype == bool and other.vec.dtype == bool, "Vector types should be the same."
        return type(self)(self.vec | other.vec, self.scheme)


class Formatter(metaclass=ABCMeta):
    @staticmethod
    @abstractmethod
    def format(code: str) -> str:
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def deformat(code: str) -> str:
        raise NotImplementedError

    @classmethod
    def reformat(cls, code: str) -> str:
        return cls.format(cls.deformat(code))


@dataclass
class LegacyUpdater(metaclass=ABCMeta):
    legacy_map: FrozenDict11[str]

    def __init__(self, legacy_map: FrozenDict11[str]):
        self.legacy_map = legacy_map

    def update(self, code: str) -> str:
        if code not in self.legacy_map:
            return code
        return self.legacy_map[code]


class CodingScheme(AbstractVxData):
    name: str
    codes: tuple[str, ...]
    desc: FrozenDict11[str]

    def __init__(self, name: str, codes: tuple[str, ...], desc: FrozenDict11[str] | None = None):
        self.name = name
        self.desc = desc or FrozenDict11({c: c for c in codes})
        self.codes = codes

    def __repr__(self):
        return f"{self.__class__.__name__}({self.name}, codes({len(self.codes)}), desc({len(self.desc)}))"

    def __check_init__(self):
        self._check_types()

        # Check types.
        assert isinstance(self.name, str), "Scheme name must be a string."
        assert isinstance(self.codes, tuple), "Scheme codes must be a tuple."
        assert all(type(c) is str for c in self.codes), "Scheme codes must be str."
        assert isinstance(self.desc, FrozenDict11), "Scheme description must be a FrozenDict11."

        assert tuple(sorted(self.codes)) == self.codes, "Scheme codes must be sorted."

        # Check for uniqueness of codes.
        assert len(self.codes) == len(set(self.codes)), f"{self}: Codes should be unique."

        # Check sizes.
        assert len(self.codes) == len(self.desc), f"{self}: Codes and descriptions should have the same size."

    def _check_types(self):
        for collection in [self.codes, self.desc]:
            assert all(isinstance(c, str) for c in collection), f"{self}: All name types should be str."

        assert all(isinstance(desc, str) for desc in self.desc.values()), f"{self}: All desc types should be str."

    @cached_property
    def index(self) -> dict[str, int]:
        return {code: idx for idx, code in enumerate(self.codes)}

    @cached_property
    def index2code(self) -> dict[int, str]:
        return {idx: code for code, idx in self.index.items()}

    @cached_property
    def index2desc(self) -> dict[int, str]:
        return {self.index[code]: _desc for code, _desc in self.desc.items()}

    def __len__(self) -> int:
        """
        Returns the number of codes in the current scheme.
        """
        return len(self.codes)

    def __bool__(self) -> bool:
        """
        Returns True if the current scheme is not empty.
        """
        return len(self.codes) > 0

    def __str__(self) -> str:
        """
        Returns the name of the current scheme.
        """
        return self.name

    def __contains__(self, code: str) -> bool:
        """Returns True if `code` is contained in the current scheme."""
        return code in self.codes

    def search_regex(self, query: str, regex_flags: int = re.I) -> set[str]:
        """
        A regex-supported search of codes by a `query` string. the search is applied on the code description.
        For example, you can use it to return all codes related to cancer by setting the
        `query = 'cancer'` and `regex_flags = re.i` (for case-insensitive search).

        Args:
            query (str): the query string.
            regex_flags (int): the regex flags.

        Returns:
            set[str]: the set of codes matching the query.
        """
        return set(filter(lambda c: re.findall(query, self.desc[c], flags=regex_flags), self.codes))

    def codeset2vec(self, codeset: Iterable[str]) -> CodesVector:
        """
        Convert a codeset to a vector representation.
        Args:
            codeset (set[str]): the codeset to convert.
        Returns:
            CodesVector: a vector representation of the current scheme.
        """
        vec = np.zeros(len(self), dtype=bool)
        try:
            for c in codeset:
                vec[self.index[c]] = True
        except KeyError as missing:
            logging.error(f"Code {missing} is missing.Accepted keys: {self.index.keys()}")

        return CodesVector(vec, self.name)

    def vec2codeset(self, vec: Array) -> set[str]:
        assert vec.ndim == 1, "Vector should be 1-dimensional."
        assert len(vec) == len(self), f"Vector length should be {len(self)}."
        return set(map(lambda idx: self.index2code[int(idx)], np.where(vec)[0]))

    def as_dataframe(self, codes: Collection[str] | None = None) -> pd.DataFrame:
        """
        Returns the scheme as a Pandas DataFrame.
        The DataFrame contains the following columns:
            - code: the code string
            - desc: the code description
        """
        if codes is None:
            codes = self.codes
        index = pd.Series(list(map(self.index.get, codes)))
        return pd.DataFrame(
            {
                "code": self.index2code,
                "desc": self.index2desc,
            },
            index=index,
        )

    @classmethod
    def _init_args_from_table(
        cls,
        name: str,
        table: pd.DataFrame,
        c_code: str,
        c_desc: str | None,
        code_selection: pd.DataFrame | None,
        **kwargs,
    ) -> tuple[str, tuple[str, ...], FrozenDict11]:
        # TODO: test this method.
        # drop=False in case we use c_code as c_desc.
        df = table.astype({c_code: str}).drop_duplicates(c_code).set_index(c_code, drop=False)
        if code_selection is not None:
            df = df.loc[code_selection[c_code].drop_duplicates().astype(str).tolist()]
        return name, tuple(sorted(df.index.tolist())), FrozenDict11(df[(c_desc or c_code)].to_dict())

    @classmethod
    def from_table(
        cls,
        name: str,
        table: pd.DataFrame,
        c_code: str,
        c_desc: str | None = None,
        code_selection: pd.DataFrame | None = None,
        *args,
        **kwargs,
    ) -> Self:
        return cls(
            *cls._init_args_from_table(
                name=name, table=table, c_code=c_code, c_desc=c_desc, code_selection=code_selection
            )
        )


class NumericScheme(CodingScheme):
    """
    NumericScheme is a subclass of FlatScheme that represents a numerical coding scheme.
    Additional to `FlatScheme` attributes, it contains the following attributes to represent the coding scheme:
    - type_hint: dict mapping codes to their type hint (B: binary, N: numerical, O: ordinal, C: categorical)
    """

    type_hint: FrozenDict11[NumericalTypeHint]
    default_type_hint: NumericalTypeHint
    group: FrozenDict11[str]

    def __init__(
        self,
        name: str,
        codes: tuple[str, ...],
        desc: FrozenDict11[str] | None = None,
        group: FrozenDict11[str] | None = None,
        type_hint: FrozenDict11[NumericalTypeHint] | None = None,
        default_type_hint: NumericalTypeHint = "N",
    ):
        super().__init__(name=name, codes=codes, desc=desc)
        self.group = group or FrozenDict11({code: code for code in codes})
        self.type_hint = type_hint or FrozenDict11({code: default_type_hint for code in codes})
        self.default_type_hint = default_type_hint

    def __check_init__(self):
        assert set(self.codes) == set(self.type_hint.keys()), (
            f"The set of codes ({self.codes}) does not match the set of type hints ({self.type_hint.keys()})."
        )
        assert set(self.type_hint.values()) <= {"B", "N", "O", "C"}, (
            f"The set of type hints ({self.type_hint.values()}) contains invalid values."
        )

    @cached_property
    def types(self) -> tuple[NumericalTypeHint, ...]:
        """
        Returns the type hint of the codes in the scheme as a numpy array.
        """
        assert set(self.index[c] for c in self.codes) == set(range(len(self))), (
            f"The order of codes ({self.codes}) does not match the order of type hints ({self.type_hint.keys()})."
        )
        return tuple(self.type_hint[code] for code in self.codes)

    @cached_property
    def index2group(self) -> dict[int, str]:
        return {i: self.group[code] for i, code in enumerate(self.codes)}

    def as_dataframe(self, codes: Collection[str] | None = None) -> pd.DataFrame:
        if codes is None:
            codes = self.codes
        index = pd.Series(list(map(self.index.get, codes)))
        return pd.DataFrame(
            {"code": self.index2code, "desc": self.index2desc, "type": self.types, "group": self.index2group},
            index=index,
        )


class CodingSchemeWithUOM(CodingScheme):
    uom_normalization_factor: FrozenDict1NM  # dict[code, dict[unit, conversion_factor]]
    universal_unit: FrozenDict11

    def __init__(
        self,
        name: str,
        codes: tuple[str, ...],
        desc: FrozenDict11,
        uom_normalization_factor: FrozenDict1NM,
        universal_unit: FrozenDict11,
    ):
        super().__init__(name=name, codes=codes, desc=desc)
        self.uom_normalization_factor = uom_normalization_factor
        self.universal_unit = universal_unit

    @classmethod
    def from_table(
        cls,
        name: str,
        table: pd.DataFrame,
        c_code: str,
        c_desc: str | None = None,
        code_selection: pd.DataFrame | None = None,
        *,
        c_normalization_factor: str,
        c_unit: str,
        c_universal_unit: str | None = None,
    ) -> Self:
        name, codes, desc = cls._init_args_from_table(
            name=name, table=table, c_code=c_code, c_desc=c_desc, code_selection=code_selection
        )
        # TODO: test this method.
        df = table.astype({c_code: str, c_normalization_factor: float})
        df = df.assign(**{c_unit: df[c_unit].str.lower()})
        df = df.drop_duplicates([c_code, c_unit]).set_index(c_code)
        df = df[df.index.isin(codes)]
        assert all(c in df.columns for c in (c_unit, c_normalization_factor)), "Some columns are missing."
        if c_universal_unit is not None and c_universal_unit in df.columns:
            uom_universal = df.loc[:, c_universal_unit].to_dict()
        else:
            _u = df.loc[:, c_unit].to_dict()
            # 1. Choose one of the units with 1.0 as a normalization factor.
            uom_universal_a = df.loc[df[c_normalization_factor] == 1.0, c_unit].to_dict()
            # 2. Or if there is only one unit per code.
            uom_universal_b = {c: _u[c] for c, count in df.index.value_counts().to_dict().items() if count == 1}
            # 3. If any item remains, choose a universal unit without any specific rule, e.g. what remains in _u.
            uom_universal_c = {
                c: u for c, u in _u.items() if c not in (set(uom_universal_a.keys()) | set(uom_universal_b.keys()))
            }

            uom_universal = uom_universal_a | uom_universal_b | uom_universal_c

        # Narrow down the codes to those who have at least one universal unit (the target unit to which all
        # units are converted).
        df = df.loc[df.index.isin(uom_universal.keys()), :]
        uom_data = {
            str(code): code_df.set_index(c_unit)[c_normalization_factor].to_dict()
            for code, code_df in df.groupby(df.index)
        }
        return cls(
            name=name,
            codes=codes,
            desc=desc,
            uom_normalization_factor=FrozenDict1NM(uom_data),
            universal_unit=FrozenDict11(uom_universal),
        )


class HierarchicalScheme(CodingScheme):
    """
    A class representing a hierarchical coding scheme.

    This class extends the functionality of the FlatScheme class and provides
    additional methods for working with hierarchical coding schemes.
    """

    ch2pt: FrozenDict1N[str]
    dag_codes: tuple[str, ...]
    dag_desc: FrozenDict11[str]
    code2dag: FrozenDict11[str]

    def __init__(
        self,
        name: str,
        codes: tuple[str, ...],
        desc: FrozenDict11[str] | None = None,
        dag_codes: Iterable[str] | None = None,
        dag_desc: FrozenDict11[str] | None = None,
        code2dag: FrozenDict11[str] | None = None,
        *,
        ch2pt: FrozenDict1N[str],
    ):
        CodingScheme.__init__(self, name, codes, desc)
        self.ch2pt = ch2pt
        self.dag_codes = tuple(dag_codes or self.codes)
        self.dag_desc = dag_desc or self.desc
        self.code2dag = code2dag or FrozenDict11({c: c for c in self.codes})

    def __check_init__(self):
        # Check types
        assert isinstance(self.dag_codes, tuple), f"{self.dag_codes}: codes should be a list."
        assert isinstance(self.dag_desc, FrozenDict11), f"{self.dag_desc}: desc should be a dict."
        assert isinstance(self.code2dag, FrozenDict11), f"{self.code2dag}: code2dag should be a dict."
        assert isinstance(self.ch2pt, FrozenDict1N), f"{self.ch2pt}: ch2pt should be a dict."
        for collection in [
            self.dag_codes,
            self.dag_desc.values(),
            self.dag_desc.keys(),
            self.code2dag.keys(),
            self.dag_desc.values(),
            self.code2dag.values(),
            self.ch2pt.keys(),
            frozenset(v for vs in self.ch2pt.values() for v in vs),
        ]:
            assert all(isinstance(c, str) for c in collection), f"{self}: All name types should be str."

        # Check sizes
        # TODO: note in the documentation that dag2code size can be less than the dag_codes since some dag_codes are
        #  internal nodes that themselves are not are not complete clinical concepts.
        for collection in [self.dag_codes, self.dag_desc]:
            assert len(collection) == len(self.dag_codes), f"{self}: All collections should have the same size."

    def make_ancestors_tuples(self, include_itself: bool = True) -> tuple[tuple[int, ...], ...]:
        """
        Creates a list of ancestors for each code.

        Args:
            include_itself (bool): whether to include the code itself as its own ancestor. Defaults to True.

        Returns:
            Array: a boolean matrix where each element (i, j) is True if code i is an ancestor of code j, and
                False otherwise.
        """
        parents_indices = [[] for _ in range(len(self.dag_index))]
        for code_i, i in self.dag_index.items():
            for ancestor_j in self.code_ancestors_bfs(code_i, include_itself):
                parents_indices[i].append(self.dag_index[ancestor_j])
        return tuple(tuple(e) for e in parents_indices)

    @cached_property
    def dag_index(self) -> dict[str, int]:
        """
        dict[str, int]: a dictionary mapping codes to their indices in the hierarchy.
        """
        return {c: i for i, c in enumerate(self.dag_codes)}

    @cached_property
    def dag2code(self) -> dict[str, str]:
        """
        dict[str, str]: a dictionary mapping codes in the hierarchy to their corresponding codes.
        """
        return {d: c for c, d in self.code2dag.items()}

    @cached_property
    def pt2ch(self) -> FrozenDict1N[str]:
        return self.reverse_connection(self.ch2pt.data)

    def __contains__(self, code: str) -> bool:
        """
        Checks if a code is contained in the current hierarchy.

        Args:
            code (str): the code to check.

        Returns:
            bool: true if the code is contained in the hierarchy, False otherwise.
        """
        return code in self.dag_codes or code in self.codes

    @staticmethod
    def reverse_connection(connection: Mapping[str, frozenset[str]]) -> FrozenDict1N[str]:
        """
        Reverses a connection dictionary.

        Args:
            connection (dict[str, set[str]]): the connection dictionary to reverse.

        Returns:
            dict[str, set[str]]: the reversed connection dictionary.
        """
        rev_connection: dict[str, set[str]] = defaultdict(set)
        for node, conns in connection.items():
            for conn in conns:
                rev_connection[conn].add(node)
        return FrozenDict1N(rev_connection)

    @staticmethod
    def _bfs_traversal(connection: Mapping[str, frozenset[str]], code: str, include_itself: bool) -> list[str]:
        """
        Performs a breadth-first traversal of the hierarchy.

        Args:
            connection (dict[str, set[str]]): the connection dictionary representing the hierarchy.
            code (str): the starting code for the traversal.
            include_itself (bool): whether to include the starting code in the traversal.

        Returns:
            list[str]: a list of codes visited during the traversal.
        """
        result = OrderedDict()
        q = [code]

        while len(q) != 0:
            # remove the first element from the stack
            current_code = q.pop(0)
            current_connections = connection.get(current_code) or ()
            q.extend([c for c in current_connections if c not in result])
            if current_code not in result:
                result[current_code] = 1

        if not include_itself:
            del result[code]
        return list(result.keys())

    @staticmethod
    def _dfs_traversal(connection: FrozenDict1N, code: str, include_itself: bool) -> list[str]:
        """
        Performs a depth-first traversal of the hierarchy.

        Args:
            connection (dict[str, set[str]]): the connection dictionary representing the hierarchy.
            code (str): the starting code for the traversal.
            include_itself (bool): whether to include the starting code in the traversal.

        Returns:
            list[str]: A list of codes visited during the traversal.
        """
        result = [code] if include_itself else []

        def _traversal(_node):
            for conn in connection.get(_node) or ():
                result.append(conn)
                _traversal(conn)

        _traversal(code)

        return list(set(result))

    @staticmethod
    def _dfs_edges(connection: FrozenDict1N, code: str) -> set[tuple[str, str]]:
        """
        Returns the edges of the hierarchy obtained through a depth-first traversal.

        Args:
            connection (dict[str, set[str]]): the connection dictionary representing the hierarchy.
            code (str): the starting code for the traversal.

        Returns:
            set[tuple[str, str]]: a set of edges in the hierarchy.
        """
        result = []

        def _edges(_node):
            connections = connection.get(_node) or ()
            for conn in connections:
                result.append((_node, conn))
                _edges(conn)

        _edges(code)
        return set(result)

    def code_ancestors_bfs(self, code: str, include_itself: bool) -> list[str]:
        """
        Returns the ancestors of a code in the hierarchy using breadth-first traversal.

        Args:
            code (str): the code for which to find the ancestors.
            include_itself (bool): whether to include the code itself as its own ancestor. Defaults to True.

        Returns:
            list[str]: A list of ancestor codes.
        """
        return self._bfs_traversal(self.ch2pt, code, include_itself)

    def code_ancestors_dfs(self, code: str, include_itself: bool) -> list[str]:
        """
        Returns the ancestors of a code in the hierarchy using depth-first traversal.

        Args:
            code (str): the code for which to find the ancestors.
            include_itself (bool): whether to include the code itself as its own ancestor. Defaults to True.

        Returns:
            list[str]: a list of ancestor codes.
        """
        return self._dfs_traversal(self.ch2pt, code, include_itself)

    def code_successors_bfs(self, code: str, include_itself: bool) -> list[str]:
        """
        Returns the successors of a code in the hierarchy using breadth-first traversal.

        Args:
            code (str): the code for which to find the successors.
            include_itself (bool): whether to include the code itself as its own successor. Defaults to True.

        Returns:
            list[str]: A list of successor codes.
        """
        return self._bfs_traversal(self.pt2ch, code, include_itself)

    def code_successors_dfs(self, code: str, include_itself: bool) -> list[str]:
        """
        Returns the successors of a code in the hierarchy using depth-first traversal.

        Args:
            code (str): the code for which to find the successors.
            include_itself (bool): whether to include the code itself as its own successor. Defaults to True.

        Returns:
            list[str]: a list of successor codes.
        """
        return self._dfs_traversal(self.pt2ch, code, include_itself)

    def ancestors_edges_dfs(self, code: str) -> set[tuple[str, str]]:
        """
        Returns the edges of the hierarchy obtained through a depth-first traversal of ancestors.

        Args:
            code (str): the code for which to find the ancestor edges.

        Returns:
            set[tuple[str, str]]: a set of edges in the hierarchy.
        """
        return self._dfs_edges(self.ch2pt, code)

    def successors_edges_dfs(self, code: str) -> set[tuple[str, str]]:
        """
        Returns the edges of the hierarchy obtained through a depth-first traversal of successors.

        Args:
            code (str): the code for which to find the successor edges.

        Returns:
            set[tuple[str, str]]: a set of edges in the hierarchy.
        """
        return self._dfs_edges(self.pt2ch, code)

    def least_common_ancestor(self, codes: list[str]) -> str:
        """
        Finds the least common ancestor of a list of codes in the hierarchy.

        Args:
            codes (list[str]): the list of codes for which to find the least common ancestor.

        Returns:
            str: the least common ancestor code.

        Raises:
            RuntimeError: if a common ancestor is not found.
        """
        while len(codes) > 1:
            a, b = codes[:2]
            a_ancestors = self.code_ancestors_bfs(a, True)
            b_ancestors = self.code_ancestors_bfs(b, True)
            last_len = len(codes)
            for ancestor in a_ancestors:
                if ancestor in b_ancestors:
                    codes = [ancestor] + codes[2:]
            if len(codes) == last_len:
                raise RuntimeError("Common Ancestor not Found!")
        return codes[0]

    def search_regex(self, query: str, regex_flags: int = re.I) -> set[str]:
        """
        A regex-based search of codes by a `query` string. the search is \
            applied on the code descriptions. for example, you can use it \
            to return all codes related to cancer by setting the \
            `query = 'cancer'` and `regex_flags = re.i` \
            (for case-insensitive search). For all found codes, \
            their successor codes are also returned in the resutls.

        Args:
            query (str): The regex query string.
            regex_flags (int): The flags to use for the regex search. Defaults to re.I (case-insensitive).

        Returns:
            set[str]: A set of codes that match the regex query, including their successor codes.
        """
        codes = [
            len(re.findall(query, self.desc[c], flags=regex_flags)) > 0
            for c in tqdm_constructor(self.codes, leave=False)
        ]
        codes = [c for c, b in zip(self.codes, codes) if b]

        dag_codes = [
            len(re.findall(query, self.dag_desc[c], flags=regex_flags)) > 0
            for c in tqdm_constructor(self.dag_codes, leave=False)
        ]
        dag_codes = [c for c, b in zip(self.dag_codes, dag_codes) if b]

        all_codes = set(map(lambda c: self.code2dag[c], codes)) | set(dag_codes)

        for c in tqdm_constructor(list(all_codes), leave=False):
            all_codes.update(self.code_successors_dfs(c, include_itself=False))

        return all_codes


class Ethnicity(CodingScheme):
    pass


class UnsupportedMapping(ValueError):
    pass


class CodeMap(AbstractVxData):
    source_name: str
    target_name: str
    data: FrozenDict1N

    def __init__(self, source_name: str, target_name: str, data: FrozenDict1N):
        self.source_name = source_name
        self.target_name = target_name
        self.data = data

    def __check_init__(self):
        assert isinstance(self.data, FrozenDict1N), "Data should be a FrozenDict1N."

    def mapped_to_dag_space(self, target_scheme: CodingScheme) -> bool:
        """
        Returns True if the CodeMap is mapped to DAG space, False otherwise.

        Returns:
            bool: True if the CodeMap is mapped to DAG space, False otherwise.
        """
        assert target_scheme.name == self.target_name, "The target scheme must be the same as the target name."
        if not isinstance(target_scheme, HierarchicalScheme) or target_scheme.dag_codes is target_scheme.codes:
            return False
        map_target_codes = frozenset([v for vs in self.data.values() for v in vs])
        target_codes = set(target_scheme.codes)
        target_dag_codes = set(target_scheme.dag_codes)
        is_code_subset = map_target_codes.issubset(target_codes)
        is_dag_subset = map_target_codes.issubset(target_dag_codes)
        assert is_code_subset or is_dag_subset, (
            "The target codes are not a subset of the target codes or the DAG codes."
        )
        return is_dag_subset

    def support_ratio(self, source_scheme: CodingScheme) -> float:
        """
        Returns the ratio between the source scheme codes covered by the mapping and the total source scheme codes.

        Returns:
            float: the support ratio of the CodeMap.
        """
        assert self.source_name == source_scheme.name, "The source scheme must be the same as the source name."
        assert self.domain.issubset(source_scheme.codes)
        return len(self.domain) / len(source_scheme.codes)

    def range_ratio(self, target_scheme: CodingScheme) -> float:
        """
        Returns the ratio between the target scheme codes covered by the mapping and the total target scheme codes.

        Returns:
            float: the range ratio of the CodeMap.
        """
        assert self.target_name == target_scheme.name, "The target scheme must be the same as the target name."
        assert self.range.issubset(target_scheme.codes)
        return len(self.range) / len(target_scheme.codes)

    def log_ratios(self, source_scheme: CodingScheme, target_scheme: CodingScheme) -> float:
        """
        Returns the score of mapping between schemes as the logarithm of support ratio multiplied by range ratio.
        """
        return math.log(self.support_ratio(source_scheme)) + math.log(self.range_ratio(target_scheme))

    def __str__(self):
        """
        Returns a string representation of the CodeMap.

        Returns:
            str: the string representation of the CodeMap.
        """
        return f"{self.source_name}->{self.target_name}"

    def __repr__(self):
        return f"{self.__class__.__name__}({self.source_name}->{self.target_name}, {len(self.data)} mappings)"

    def __hash__(self):
        """
        Returns the hash value of the CodeMap.

        Returns:
            int: the hash value of the CodeMap.
        """
        return hash(str(self))

    def __bool__(self):
        """
        Returns True if the CodeMap is not empty, False otherwise.

        Returns:
            bool: true if the CodeMap is not empty, False otherwise.
        """
        return len(self) > 0

    def __len__(self):
        """
        Returns the number of supported codes in the CodeMap.

        Returns:
            int: the number of supported codes in the CodeMap.
        """
        return len(self.data)

    def target_index(self, target_scheme: CodingScheme | HierarchicalScheme) -> dict[str, int]:
        """
        Returns the target coding scheme index.

        Returns:
            dict: the target coding scheme index.
        """
        assert self.target_name == target_scheme.name, "The target scheme must be the same as the target name."
        if (
            isinstance(target_scheme, HierarchicalScheme)
            and self.mapped_to_dag_space(target_scheme)
            and self.source_name != self.target_name
            and hasattr(target_scheme, "dag_index")
        ):
            return target_scheme.dag_index
        return target_scheme.index

    def target_desc(self, target_scheme: CodingScheme | HierarchicalScheme):
        """
        Returns the target coding scheme description.

        Returns:
            dict: the target coding scheme description.
        """
        assert self.target_name == target_scheme.name, "The target scheme must be the same as the target name."
        if (
            isinstance(target_scheme, HierarchicalScheme)
            and self.mapped_to_dag_space(target_scheme)
            and self.source_name != self.target_name
            and hasattr(target_scheme, "dag_desc")
        ):
            return target_scheme.dag_desc
        return target_scheme.desc

    def __getitem__(self, item):
        """
        Returns the mapped codes for the given item.

        Args:
            item: the item to retrieve the mapped codes for.

        Returns:
            set[str]: the mapped codes for the given item.
        """
        return self.data[item]

    def __contains__(self, item):
        """
        Checks if an item is in the CodeMap.

        Args:
            item: the item to check.

        Returns:
            bool: True if the item is in the CodeMap, False otherwise.
        """
        return item in self.data

    @cached_property
    def domain(self) -> frozenset[str]:
        return frozenset(self.data.keys())

    @cached_property
    def range(self) -> frozenset[str]:
        return frozenset(v for vs in self.data.values() for v in vs)

    def map_codeset(self, codeset: Iterable[str]) -> frozenset:
        """
        Maps a codeset to the target coding scheme.

        Args:
            codeset (set[str]): the codeset to map.

        Returns:
            set[str]: The mapped codeset.
        """
        supported_set = self.domain.intersection(codeset)
        if len(supported_set) == 0 and len(cast(Sized, codeset)) > 0:
            logging.debug(f"No code in ({codeset}) maps to the target coding scheme.")
            return frozenset()
        return frozenset(t for c in supported_set for t in self[c])

    def map_dataframe(self, df: pd.DataFrame, code_column: str) -> pd.DataFrame:
        df = df.iloc[:, :]
        code2list = {k: list(v) if len(v) > 1 else next(iter(v)) for k, v in self.data.items()}
        df[code_column] = df.loc[:, code_column].map(code2list)
        invalid_rows = df[code_column].isna()
        if invalid_rows.sum() > 0:
            unique_codes = df.loc[invalid_rows, code_column].unique()
            total_unique_codes = df[code_column].unique()
            dataframe_log.info(
                (
                    f"Some codes are not mapped to the target scheme: "
                    f"Total rows removed {invalid_rows.sum()} / {len(invalid_rows)} = {invalid_rows.mean(): .3f}. "
                    f"Unique codes dropped: {len(unique_codes)} / {len(total_unique_codes)} = "
                    f"{len(unique_codes) / len(total_unique_codes): .3f}.",
                    pd.DataFrame(unique_codes, columns=pd.Series([code_column])),
                    "unique_columns_missed",
                )
            )
            df = df.loc[~df[code_column].isna(), :]
        return df.explode(code_column)

    def target_code_ancestors(self, target_scheme: HierarchicalScheme, t_code: str, include_itself=True) -> list[str]:
        assert self.target_name == target_scheme.name, "The target scheme must be the same as the target name."
        if not self.mapped_to_dag_space(target_scheme):
            t_code = target_scheme.code2dag[t_code]
        return target_scheme.code_ancestors_bfs(t_code, include_itself=include_itself)

    @classmethod
    def _init_args_from_table(
        cls,
        source_scheme: CodingScheme,
        target_scheme: CodingScheme,
        map_table: pd.DataFrame,
        c_source_code: str,
        c_target_code: str,
        *args,
        **kwargs,
    ) -> tuple[str, str, FrozenDict1N]:
        """
        # TODO: test me.
        """
        map_table = map_table.loc[:, [c_source_code, c_target_code]].astype(str)
        map_table = map_table.loc[
            map_table[c_source_code].isin(source_scheme.codes) & map_table[c_target_code].isin(target_scheme.codes), :
        ]
        mapping = map_table.groupby(c_source_code)[c_target_code].apply(set).to_dict()
        return source_scheme.name, target_scheme.name, FrozenDict1N(mapping)

    @classmethod
    def from_table(
        cls,
        source_scheme: CodingScheme,
        target_scheme: CodingScheme,
        c_source_code: str,
        c_target_code: str,
        table: pd.DataFrame,
        **kwargs,
    ):
        return cls(
            *cls._init_args_from_table(
                source_scheme=source_scheme,
                target_scheme=target_scheme,
                c_source_code=c_source_code,
                c_target_code=c_target_code,
                map_table=table,
            )
        )


class GroupingData(AbstractVxData):
    permute: tuple[int, ...]
    split: tuple[int, ...]
    size: tuple[int, ...]
    aggregation: tuple[AggregationLiteral, ...]

    def __init__(
        self,
        permute: tuple[int, ...],
        split: tuple[int, ...],
        size: tuple[int, ...],
        aggregation: tuple[AggregationLiteral, ...],
    ):
        self.permute = permute
        self.split = split
        self.size = size
        self.aggregation = aggregation

    @property
    def scheme_size(self) -> tuple[int, int]:
        return sum(self.size), len(self.size)


class ReducedCodeMapN1(CodeMap):
    set_aggregation: FrozenDict11
    reduced_groups: FrozenDict1N

    def __init__(
        self,
        source_name: str,
        target_name: str,
        data: FrozenDict1N,
        set_aggregation: FrozenDict11,
        reduced_groups: FrozenDict1N,
    ):
        super().__init__(source_name=source_name, target_name=target_name, data=data)
        self.set_aggregation = set_aggregation
        self.reduced_groups = reduced_groups

    @classmethod
    def from_data(
        cls, source_name: str, target_name: str, map_data: FrozenDict1N, set_aggregation: FrozenDict11
    ) -> Self:
        assert all(len(t) == 1 for t in map_data.values()), "A code should have one target."

        new_map: dict[str, set[str]] = defaultdict(set)
        for code, (target,) in map_data.items():
            new_map[target].add(code)

        return cls(
            source_name=source_name,
            target_name=target_name,
            data=map_data,
            set_aggregation=set_aggregation,
            reduced_groups=FrozenDict1N(new_map),
        )

    @classmethod
    def from_table(
        cls,
        source_scheme: CodingScheme,
        target_scheme: CodingScheme,
        c_source_code: str,
        c_target_code: str,
        table: pd.DataFrame,
        **kwargs,
    ) -> Self:
        c_target_agg = kwargs["c_target_agg"]
        source_name, target_name, map_data = cls._init_args_from_table(
            source_scheme=source_scheme,
            target_scheme=target_scheme,
            c_source_code=c_source_code,
            c_target_code=c_target_code,
            map_table=table,
        )
        return cls.from_data(
            source_name=source_name,
            target_name=target_name,
            map_data=map_data,
            set_aggregation=FrozenDict11(table.set_index(c_target_code)[c_target_agg].to_dict()),
        )

    def groups(self, source_index: dict[str, int]) -> tuple[tuple[str, ...], ...]:
        target_codes = tuple(sorted(self.reduced_groups.keys()))
        return tuple(tuple(sorted(self.reduced_groups[t], key=lambda c: source_index[c])) for t in target_codes)

    @staticmethod
    def _validate_aggregation(a) -> AggregationLiteral:
        if a in ("sum", "or", "w_sum"):
            return a
        else:
            raise ValueError(f"Unrecognised aggregation: {a}")

    @cached_property
    def groups_aggregation(self) -> tuple[AggregationLiteral, ...]:
        aggregation = tuple(self.set_aggregation[g] for g in sorted(self.reduced_groups.keys()))
        return tuple(self._validate_aggregation(a) for a in aggregation)

    def groups_size(self, source_index: dict[str, int]) -> tuple[int, ...]:
        return tuple(len(g) for g in self.groups(source_index))

    def groups_split(self, source_index: dict[str, int]) -> tuple[int, ...]:
        return tuple(np.cumsum(self.groups_size(source_index)).tolist())

    def groups_permute(self, source_index: dict[str, int]) -> tuple[int, ...]:
        permutes: tuple[int, ...] = sum(
            (tuple(map(source_index.__getitem__, g)) for g in self.groups(source_index)), tuple()
        )
        if len(permutes) == len(source_index):
            return permutes
        else:
            return permutes + tuple(set(source_index.values()) - set(permutes))

    def grouping_data(self, source_index: dict[str, int]) -> GroupingData:
        return GroupingData(
            permute=self.groups_permute(source_index),
            split=self.groups_split(source_index),
            size=self.groups_size(source_index),
            aggregation=self.groups_aggregation,
        )


class FilterOutcomeMapData(AbstractVxData):
    name: str
    base_name: str
    exclude_codes: tuple[str, ...]

    def __init__(self, name: str, exclude_codes: tuple[str, ...], base_name: str):
        self.name = name
        self.exclude_codes = exclude_codes
        self.base_name = base_name

    @classmethod
    def from_spec_json(cls, available_schemes: Mapping[str, CodingScheme], json_file: str) -> Self:
        conf = load_config(json_file, relative_to=resources_path("outcomes"))
        exclude_codes = []
        if "exclude_branches" in conf:
            pass
        if "select_branches" in conf:
            pass
        if "selected_codes" in conf:
            base_scheme = available_schemes[conf["code_scheme"]]
            exclude_codes.extend(c for c in base_scheme.codes if c not in conf["selected_codes"])
        if "exclude_codes" in conf:
            exclude_codes.extend(conf["exclude_codes"])

        name = conf.get("name", json_file.split(".")[0])
        return cls(name=name, base_name=conf["code_scheme"], exclude_codes=tuple(exclude_codes))

    def as_coding_scheme(self, base_scheme: CodingScheme) -> CodingScheme:
        assert base_scheme.name == self.base_name, "Base scheme mismatch."
        codes = tuple(c for c in base_scheme.codes if c not in self.exclude_codes)
        desc = FrozenDict11({c: base_scheme.desc[c] for c in codes})
        return CodingScheme(name=self.name, codes=codes, desc=desc)

    def as_code_map(
        self, outcome_scheme: CodingScheme, source_scheme: CodingScheme, base_scheme: CodingScheme, source2base: CodeMap
    ) -> CodeMap:
        assert source2base.source_name == source_scheme.name
        assert source2base.target_name == base_scheme.name == self.base_name, "Base scheme mismatch."
        # as_scheme = self.as_coding_scheme(base_scheme)
        map_data = {s_code: t_codes.intersection(outcome_scheme.codes) for s_code, t_codes in source2base.data.items()}
        return CodeMap(source_name=source_scheme.name, target_name=self.name, data=FrozenDict1N(map_data))


@dataclass
class FilterOutcomeMap:
    scheme: CodingScheme
    codemap: CodeMap

    def __init__(self, scheme: CodingScheme, codemap: CodeMap):
        self.scheme = scheme
        self.codemap = codemap

    @property
    def name(self) -> str:
        return self.scheme.name

    @property
    def index(self):
        return self.scheme.index

    @property
    def desc(self):
        return self.scheme.desc

    def __len__(self):
        return len(self.index)

    def map_codeset(self, codeset: Iterable[str]):
        return self.codemap.map_codeset(codeset)

    def __call__(self, codeset: Iterable[str]) -> CodesVector:
        codeset = self.map_codeset(codeset)
        vec = np.zeros(len(self), dtype=bool)
        for c in codeset:
            vec[self.index[c]] = True
        return CodesVector(vec, self.name)


class CodingSchemesManager(AbstractVxData):
    schemes: tuple[CodingScheme, ...]
    maps: tuple[CodeMap, ...]
    outcomes: tuple[FilterOutcomeMapData, ...]

    def __init__(
        self,
        schemes: tuple[CodingScheme, ...] = (),
        maps: tuple[CodeMap, ...] = (),
        outcomes: tuple[FilterOutcomeMapData, ...] = (),
    ):
        self.schemes = tuple(sorted(schemes, key=lambda s: s.name))
        self.maps = tuple(sorted(maps, key=lambda m: m.source_name + m.target_name))
        self.outcomes = tuple(sorted(outcomes, key=lambda o: o.name))

    def __repr__(self):
        return f"{self.__class__.__name__}({self.schemes}, {self.maps}, {self.outcomes})"

    def __len__(self):
        return len(self.schemes) + len(self.maps) + len(self.outcomes)

    def add_scheme(self, scheme: CodingScheme) -> Self:
        assert isinstance(scheme, CodingScheme), f"{scheme} is not a CodingScheme."
        if scheme.name in self.scheme:
            logging.warning(f"Scheme {scheme.name} already exists")
            return self
        return type(self)(schemes=self.schemes + (scheme,), maps=self.maps, outcomes=self.outcomes)

    def add_map(self, map: CodeMap, overwrite: bool = False) -> Self:
        assert isinstance(map, CodeMap), f"{map} is not a CodeMap."
        if (map.source_name, map.target_name) in self.map:
            if not overwrite:
                logging.warning(
                    f"Map {map.source_name}->{map.target_name} already exists. "
                    f"If you want to replace it, use overwrite=True."
                )
                return self
            logging.info(f"Map {map.source_name}->{map.target_name} already exists and will be overwritten")

        return type(self)(schemes=self.schemes, maps=self.maps + (map,), outcomes=self.outcomes)

    def add_outcome(self, outcome: FilterOutcomeMapData) -> Self:
        assert isinstance(outcome, FilterOutcomeMapData), f"{outcome} is not an OutcomeExtractor."
        if outcome.name in self.outcome_data:
            logging.warning(f"Outcome {outcome.name} already exists")
            return self
        return type(self)(schemes=self.schemes, maps=self.maps, outcomes=self.outcomes + (outcome,))

    def supported_outcome(self, outcome_name: str, source_scheme: str) -> bool:
        assert outcome_name in self.outcome_data, (
            f"This outcome ({outcome_name}) doesn't exist. Current outcomes: {list(self.outcome_data.keys())}"
        )

        return (source_scheme, outcome_name) in self.outcome

    def supported_outcomes(self, source_scheme: str) -> tuple[str, ...]:
        return tuple(o.name for o in self.outcomes if self.supported_outcome(o.name, source_scheme))

    def union(self, other: Self) -> Self:
        updated = self
        for s in (s for s in other.schemes if s.name not in updated.scheme):
            updated = updated.add_scheme(s)
        for m in (m for m in other.maps if (m.source_name, m.target_name) not in updated.map):
            updated = updated.add_map(m)
        for o in (o for o in other.outcomes if o.name not in updated.outcome_data):
            updated = updated.add_outcome(o)
        return updated

    def __add__(self, other: Self) -> Self:
        return self.union(other)

    @cached_property
    def scheme(self) -> Mapping[str, CodingScheme]:
        return MappingProxyType({s.name: s for s in self.schemes})

    @cached_property
    def identity_maps(self) -> dict[tuple[str, str], CodeMap]:
        return {
            (s, s): CodeMap(source_name=s, target_name=s, data=FrozenDict1N({c: {c} for c in self.scheme[s].codes}))
            for s in self.scheme.keys()
        }

    @cached_property
    def chainable_maps(self) -> dict[tuple[str, str, str], CodeMap]:
        # TODO: to automatically generate chained maps, we need to define a graph algorithm traversing
        # the paths from a source node to target node.
        # Notes:
        # 1. Maybe restrict the chained maps to only two edges. Or allow a hyperparameter (max_edges).
        # 2. There will be a need to define quality metrics of chained maps, because there are potentially
        #   multiple chained maps from node i to node j.
        #   Quality metric would consider the following factors:
        #   1. Lost codes in the mapping. If we have multiple paths from node i to node j, favor the ones that
        #       loses least number of codes of node i in the mapping.
        #   2. Spread of codes in the target. If we have multiple paths from node i to node j, favor the ones
        #       that does distributes more evenly over the codes of node j. Perhaps this can be
        #       quantified with entropy. Let
        #       p_k = (number of mapped codes from node i to code k in node j) / |normalizer|,
        #       then $entropy = \sum p_k \log{p_k}$
        # How to combine these two objectives?
        raise NotImplementedError

    @cached_property
    def map(self) -> Mapping[tuple[str, str], CodeMap]:
        return MappingProxyType({(m.source_name, m.target_name): m for m in self.maps} | self.identity_maps)

    @cached_property
    def outcome_data(self) -> Mapping[str, FilterOutcomeMapData]:
        return MappingProxyType({o.name: o for o in self.outcomes})

    @cached_property
    def outcome_scheme(self) -> Mapping[str, CodingScheme]:
        return MappingProxyType({k: o.scheme for (_, k), o in self.outcome.items()})

    @cached_property
    def outcome(self) -> Mapping[tuple[str, str], FilterOutcomeMap]:
        # get all outcome mappings possible.
        # basically includes any codemap that has as a target base_scheme. for all base_schemes.
        results = {}
        for o in self.outcomes:
            t_scheme = o.base_name
            o_scheme = o.as_coding_scheme(self.scheme[t_scheme])
            feasible_maps = {k: m for k, m in self.map.items() if m.target_name == t_scheme}
            for (source_name, target_name), m in feasible_maps.items():
                o_map = o.as_code_map(o_scheme, self.scheme[source_name], self.scheme[target_name], m)
                results[(source_name, o.name)] = FilterOutcomeMap(o_scheme, o_map)
        return MappingProxyType(results)

    def make_chained_map(self, chain: tuple[str, ...]) -> CodeMap:
        """
        Registers a chained CodeMap. The source and target coding schemes are chained together if there is an
        intermediate scheme that can act as a bridge between the two. There must be registered two CodeMaps, one that
        maps between the source and intermediate coding schemes and one that maps between the intermediate and
        target coding schemes.
        Args:
            s_scheme (str): the source coding scheme.
            inter_scheme (str): the intermediate coding scheme.
            t_scheme (str): the target coding scheme.
        """
        assert len(chain) > 2
        scheme = tuple(map(lambda n: self.scheme[n], chain))
        maps = tuple(self.map[si, sj] for si, sj in zip(chain[:-1], chain[1:]))
        assert all(len(m.range.intersection(sj.codes)) > 0 for m, sj in zip(maps, scheme[1:]))
        assert all(len(m.domain.intersection(si.codes)) > 0 for m, si in zip(maps, scheme[:-1]))

        def bridge(c: str) -> frozenset[str]:
            codeset = frozenset((c,))
            for m in maps:
                codeset = m.map_codeset(codeset)
                if len(codeset) == 0:
                    return frozenset()
            return codeset

        data = {c: bridge(c) for c in scheme[0].codes}
        data = FrozenDict1N({c: target for c, target in data.items() if len(target) > 0})
        return CodeMap(source_name=chain[0], target_name=chain[-1], data=data)

    def add_chained_map(self, s_scheme: str, inter_scheme: str, t_scheme: str, overwrite: bool = False) -> Self:
        """
        Registers a chained CodeMap. The source and target coding schemes are chained together if there is an
        intermediate scheme that can act as a bridge between the two. There must be registered two CodeMaps, one that
        maps between the source and intermediate coding schemes and one that maps between the intermediate and target
        coding schemes.
        Args:
            s_scheme (str): the source coding scheme.
            inter_scheme (str): the intermediate coding scheme.
            t_scheme (str): the target coding scheme.
        """
        return self.add_map(self.make_chained_map((s_scheme, inter_scheme, t_scheme)), overwrite=overwrite)

    def scheme_supported_targets(self, scheme: CodingScheme) -> tuple[str, ...]:
        return tuple(t for s, t in self.map.keys() if s == scheme.name)

    def add_match_map(self, scheme_a: str, scheme_b: str) -> Self:
        scheme_a_ = self.scheme[scheme_a]
        scheme_b_ = self.scheme[scheme_b]
        normalize = lambda c: c.strip().lower()
        normalized_a = tuple(sorted(map(normalize, scheme_a_.codes)))
        normalized_b = tuple(sorted(map(normalize, scheme_b_.codes)))
        assert normalized_a == normalized_b, (
            f"The codes of {scheme_a_.name} and {scheme_b_.name} mismatch. "
            f"(Normalized) codes of {scheme_a_.name}: {normalized_a}. "
            f"(Normalized) codes of {scheme_b_.name}: {normalized_b}."
            f"Difference a - b: {set(normalized_a) - set(normalized_b)}. "
            f"Difference b - a: {set(normalized_b) - set(normalized_a)}."
        )
        codes_a = tuple(sorted(scheme_a_.codes, key=normalize))
        codes_b = tuple(sorted(scheme_b_.codes, key=normalize))
        m_ab = CodeMap(scheme_a_.name, scheme_b_.name, data=FrozenDict1N({a: {b} for a, b in zip(codes_a, codes_b)}))
        m_ba = CodeMap(scheme_b_.name, scheme_a_.name, data=FrozenDict1N({b: {a} for a, b in zip(codes_a, codes_b)}))
        return self.add_map(m_ab).add_map(m_ba)



"""Utility helpers used across the project.

This module provides a small collection of utilities for:

- Resolving and translating filesystem paths
- Simple JSON configuration load/save with NumPy support
- Introspecting attribute/key access paths from callables and JAX keypaths
- Light array helpers that work with both NumPy and JAX arrays
- A logging adapter that persists `pandas.DataFrame` payloads alongside logs

The functions are intentionally lightweight and have no side-effects beyond
those documented in their docstrings.
"""

import json
import logging
import os
from collections.abc import Callable, MutableMapping
from pathlib import Path
from types import ModuleType
from typing import Any, cast

import jax
import jax.numpy as jnp
import numpy as np
import pandas as pd
from jax._src.tree_util import DictKey, FlattenedIndexKey, GetAttrKey, KeyEntry, SequenceKey
from jaxlib._jax import ArrayImpl
from tqdm import tqdm
from tqdm.notebook import tqdm as tqdm_notebook


ArrayTypes = (np.ndarray, jnp.ndarray, jax.Array, ArrayImpl)
Array = np.ndarray | jnp.ndarray | jax.Array | ArrayImpl


def _tqdm_backend():
    """Return the appropriate constructor of tqdm based on the executor
    interpreter, i.e. if it is running on a notebook or not."""
    try:
        ipy_str = str(type(get_ipython()))  # type: ignore
        if "zmqshell" in ipy_str:
            return tqdm_notebook
    except NameError:
        pass
    return tqdm


tqdm_constructor = _tqdm_backend()


def translate_path(path: str, relative_to: str | None = None):
    """Translate a filesystem path by replacing environment
    variables with their values. If relative_to is specified,
    and a path is a relative path to a nonexistent file, then the
    path is interpreted as relative to this path.

    Parameters:
        path: Filesystem path to translate.
        relative_to: if specified, and a path is a relative path to a nonexistent file, then the path is
            interpreted as relative to this path.
    Returns:
        Absolute, expanded, translated path.
    """
    assert isinstance(path, str), "Path must be a string."

    path = os.path.expandvars(os.path.expanduser(path))

    if relative_to is not None:
        relative_to = translate_path(relative_to)
    if relative_to is not None and not os.path.isabs(path) and not os.path.exists(path):
        path = os.path.join(relative_to, path)

    return os.path.abspath(path)


def resources_path(*subdir: str) -> str:
    """Return absolute path inside the package `resources` directory.

    Parameters:
        subdir: Optional path segments to append within the `resources` folder.

    Returns:
        Absolute path pointing to `ehrax/resources/` joined with the given parts.
    """
    return str(os.path.join(os.path.dirname(__file__), "resources", *subdir))


def load_config(config_file: str, relative_to: str | None = None):
    """Load a JSON file from `config_file`.

    Parameters:
        config_file: Path to the JSON configuration file to load.
        relative_to: if specified, and config_file is a relative path to a nonexistent file, then the path is
            interpreted as relative to this path.

    Returns:
        The deserialized JSON contents of the configuration file.
    """
    config_file = translate_path(config_file, relative_to=relative_to)
    with open(config_file) as json_file:
        return json.load(json_file)


def write_config(data, config_file):
    """Write the given data object to the specified config file as JSON.

    Parameters:
        data: The data object to serialize to JSON and write.
        config_file: The path to the config file to write.
    """
    with open(translate_path(config_file), "w") as outfile:
        json.dump(data, outfile, indent=4, sort_keys=True, cls=NumpyEncoder)


def path_from_getter(
    getter: Callable[[Any], Any],
    getattr_transform: Callable[[str], str] = lambda x: x,
    getitem_transform: Callable[[Any], str] = lambda x: x,
) -> list[str]:
    """
    Generate a sequence of attribute names or indices (converted to strings) recording the sequence of access steps
    applied by the function on its input.

    !!! Example

    ```python
    path_from_getter(lambda x: x.y.z.a.b)
    # returns ['y', 'z', 'a', 'b']
    path_from_getter(lambda x: x["y"][4].money)
    # returns ['y', '4', 'money']
    ```
    """

    class _M:
        _x_path: list[str]

        def __init__(self, _x_path: list[str]):
            self._x_path = _x_path

        def __getattribute__(self, item: str):
            try:
                return object.__getattribute__(self, item)
            except AttributeError:
                return _M(object.__getattribute__(self, "_x_path") + [getattr_transform(item)])

        def __getitem__(self, item: str):
            return _M(object.__getattribute__(self, "_x_path") + [str(getitem_transform(item))])

    return getter(_M([]))._x_path


def path_from_jax_keypath(
    path: tuple[KeyEntry, ...],
    getattr_transform: Callable[[str], str] = lambda x: x,
    getitem_transform: Callable[[Any], str] = lambda x: x,
) -> list[str]:
    """Convert a JAX keypath into a list of human-readable path elements.

    Parameters:
        path: Sequence of JAX `KeyEntry` elements that describe how to reach a
            value within a pytree.
        getattr_transform: Optional mapping applied to attribute names.
        getitem_transform: Optional mapping applied to indices/keys.

    Returns:
        List of strings representing successive attribute names or indices/keys.
    """

    def _extract(entry: KeyEntry):
        match entry:
            case GetAttrKey(name):
                return getattr_transform(name)
            case SequenceKey(idx):
                return str(getitem_transform(idx))
            case DictKey(key):
                return str(getitem_transform(key))
            case FlattenedIndexKey(key):
                return str(getitem_transform(key))
            case _:
                raise ValueError(f"Unexpected key {entry}")

    return list(map(_extract, path))


class NumpyEncoder(json.JSONEncoder):
    """Custom JSON encoder for NumPy scalars and arrays.

    Encodes NumPy integers/floats/bools to the corresponding Python scalars,
    complex numbers to a mapping with ``{"real": ..., "imag": ...}``, arrays
    to lists, and ``np.void`` to ``None``. Falls back to the default encoder
    if the object is not recognized.
    """

    def default(self, obj: object) -> object:  # type: ignore
        if np.issubdtype(type(obj), np.integer):
            return int(obj)  # type: ignore
        elif np.isreal(obj):  # type: ignore
            return float(obj)  # type: ignore
        elif np.iscomplex(obj):  # type: ignore
            return {"real": obj.real, "imag": obj.imag}  # type: ignore

        elif isinstance(obj, (np.ndarray,)):
            return obj.tolist()

        elif isinstance(obj, np.bool_):
            return bool(obj)

        elif isinstance(obj, np.void):
            return None

        return json.JSONEncoder.default(self, obj)


def np_module(a: Array) -> ModuleType:  # [np, jnp]:
    """Return the numerical module (`numpy` or `jax.numpy`) for the array.

    Parameters:
        a: Array instance (NumPy ndarray or JAX array) whose module to return.

    Returns:
        The corresponding module object: either ``numpy`` or ``jax.numpy``.

    Raises:
        TypeError: If the array type is not supported.
    """
    if isinstance(a, np.ndarray):
        return np
    elif isinstance(a, jnp.ndarray):  # type: ignore
        return jnp
    else:
        raise TypeError(f"Unsupported array type {type(a)}.")


def equal_arrays(a: Array, b: Array) -> bool:
    """Compare two arrays for equality while treating NaNs as unequal values.

    Arrays must have identical shapes and dtypes. NaN positions are ignored in
    the comparison (i.e., elements where both are NaN are skipped), and all
    remaining elements must be exactly equal.

    Parameters:
        a: First array (NumPy or JAX).
        b: Second array (NumPy or JAX).

    Returns:
        True if arrays are equal under the above rules; False otherwise.
    """
    if a.shape != b.shape or a.dtype != b.dtype:
        return False
    _np = np if isinstance(a, np.ndarray) else jnp
    if _np.size(a) == 0:
        return True
    is_nan = _np.isnan(a) & _np.isnan(b)
    return _np.array_equal(a[~is_nan], b[~is_nan], equal_nan=False)  # type: ignore


class DataFrameLogger(logging.LoggerAdapter):
    """Logger adapter that writes a DataFrame to disk as a CSV on log calls.

    When used like a standard logger adapter and provided with a
    ``dataframe=pd.DataFrame`` keyword argument (and optional ``tag=str``), the
    DataFrame will be written next to the main log file using the same base
    filename plus a timestamp and incremental id. The log message is augmented
    with a short note pointing to the CSV file path.

    Note: This requires that the underlying logger has a single
    ``logging.FileHandler`` configured; otherwise, the adapter leaves the
    message unchanged and appends a hint.
    """

    @property
    def extract_file_handler_names(self) -> tuple[str, str, str] | None:
        """Return tuple of (parent_dir, file_stem, suffix) for the FileHandler.

        Returns None if the underlying logger has no single FileHandler.
        """
        for handler in self.logger.handlers:
            if isinstance(handler, logging.FileHandler):
                # baseFilename is actually the absolute path.
                # https://github.com/python/cpython/blob/801cf3fcdd27d8b6dd0fdd3c39e6c996e2b2f7fa/Lib/logging/__init__.py#L1200
                file_title = Path(handler.baseFilename).stem
                file_suffix = Path(handler.baseFilename).suffix
                file_parent = str(Path(handler.baseFilename).parent)
                return file_parent, file_title, file_suffix
        return None

    def process(self, msg: str, kwargs: MutableMapping[str, Any]) -> tuple[str, MutableMapping[str, Any]]:
        """Process a log record; persist provided DataFrame and enrich message.

        Expects the following keyword arguments in ``kwargs``:
        - ``dataframe``: a non-empty ``pandas.DataFrame`` to persist
        - ``tag`` (optional): short string to help identify the CSV file

        Returns the possibly modified message and keyword-argument mapping.
        """
        df, tag = kwargs.pop("dataframe"), kwargs.pop("tag", "")
        assert tuple(map(type, (msg, df, tag))) == (str, pd.DataFrame, str)
        if len(df) == 0:
            return msg, kwargs

        # timestamp representative and incremental id.
        timestamp = pd.Timestamp.now().strftime("%Y_%m_%dT_%H_%M_%S")
        if self.extra is None:
            self.extra = {}
        incremental_id = cast(int, self.extra.get("incremental_id", 0)) + 1
        self.extra = dict(self.extra) | {"incremental_id": incremental_id}
        filehandler_names = self.extract_file_handler_names
        if filehandler_names is None:
            return (
                f"{msg}. Appendix report will not be saved to disk because "
                f"no single file handler found in the logger. "
                f"To store the appendix report, configure the logger {self.logger.name} "
                f"by either adding a FileHandler manually or call logging.basicConfig "
                f"with setting the filename argument."
            ), kwargs
        parent_dir, main_log_file, main_log_file_suffix = filehandler_names
        tag = tag.replace(".", "_")
        file_title = "_".join((main_log_file, tag, timestamp, f"{incremental_id:03d}"))
        file_path = Path(parent_dir, file_title).with_suffix(f"{main_log_file_suffix}.csv")
        df.to_csv(file_path)
        return (
            f"{msg}. Find the appendix report stored as a table of "
            f"columns {df.columns.tolist()} and {len(df)} rows at ({file_path})."
        ), kwargs


def attached_dataframe_logger(logger: logging.Logger, extra: dict[str, Any] | None = None) -> DataFrameLogger:
    """Create a ``DataFrameLogger`` adapter for the given logger.

    Parameters:
        logger: Base logger to adapt.
        extra: Optional mapping of extra fields to include in every record.

    Returns:
        A ``DataFrameLogger`` that can be used like a regular logger adapter.
    """
    if extra is None:
        extra = {}
    return DataFrameLogger(logger, extra)


# Global logger adapter for persisting DataFrame payloads with log messages.
# Usage:
#     dataframe_log.info("Saved report", dataframe=df, tag="report")
# Ensure a FileHandler is configured on the base logger so CSVs are written
# alongside the main log file.
dataframe_log: DataFrameLogger = attached_dataframe_logger(logging.getLogger())



from typing import Self

import pandas as pd

from ..coding_scheme import CodeMap, CodingScheme, CodingSchemesManager, Formatter, FrozenDict1N, FrozenDict11
from ..dataset import COLUMN
from ..utils import dataframe_log


class MultiVersionScheme(CodingScheme):
    component_scheme_names: FrozenDict11
    sep: str

    def __init__(
        self, component_scheme_names: FrozenDict11, sep: str, *, name: str, codes: tuple[str, ...], desc: FrozenDict11
    ):
        super().__init__(name, codes, desc)
        self.component_scheme_names = component_scheme_names
        self.sep = sep

    def component_schemes(self, manager: CodingSchemesManager) -> dict[str, CodingScheme]:
        return {k: manager.scheme[v] for k, v in self.component_scheme_names.items()}

    @staticmethod
    def reformat(df: pd.DataFrame, component_schemes: dict[str, CodingScheme]) -> pd.DataFrame:
        if any(isinstance(s, Formatter) for s in component_schemes.values()):
            df = df.copy()
            id_fmt = lambda c: c
            reformat = {
                v: scheme.reformat if isinstance(scheme, Formatter) else id_fmt
                for v, scheme in component_schemes.items()
            }
            df[COLUMN.code] = list(map(lambda c, v: reformat[v](c), df[COLUMN.code], df[COLUMN.version]))
        return df

    @classmethod
    def from_selection(
        cls, name: str, multi_version_selection: pd.DataFrame | None, component_schemes: dict[str, CodingScheme]
    ) -> Self:
        if multi_version_selection is None:
            multi_version_selection = pd.concat(
                [
                    pd.DataFrame(
                        {
                            str(COLUMN.code): list(si.codes),
                            str(COLUMN.version): [v] * len(si),
                            str(COLUMN.description): list(map(si.desc.get, si.codes)),
                        }
                    )
                    for v, si in component_schemes.items()
                ]
            )
        selection = multi_version_selection.sort_values([str(COLUMN.version), str(COLUMN.code)])
        selection = selection.drop_duplicates([str(COLUMN.version), str(COLUMN.code)]).astype(str)
        assert selection[COLUMN.version].isin(component_schemes.keys()).all(), (
            f"Only versions {', '.join(f'({v}, {s.name})' for v, s in component_schemes.items())} are expected."
        )
        assert selection.groupby([str(COLUMN.version), str(COLUMN.code)]).size().max() == 1, (
            "Duplicate (version, code) pairs are not allowed."
        )

        df = cls.reformat(selection, component_schemes)
        valid_sep = lambda s: not (df[COLUMN.version].str.contains(s).any() or df[COLUMN.code].str.contains(s).any())
        sep = next(s for s in "@:-#$%!=_/+><?;|~," if valid_sep(s))
        df[COLUMN.code] = (df[COLUMN.version] + sep + df[COLUMN.code]).tolist()
        desc = df.set_index(str(COLUMN.code))[COLUMN.description].to_dict()
        return cls(
            name=name,
            codes=tuple(sorted(df[COLUMN.code].tolist())),
            desc=FrozenDict11(desc),
            component_scheme_names=FrozenDict11({k: s.name for k, s in component_schemes.items()}),
            sep=sep,
        )

    def mixed_code_format_table(self, manager: CodingSchemesManager, table: pd.DataFrame) -> pd.DataFrame:
        c_code = str(COLUMN.code)
        c_version = str(COLUMN.version)

        assert c_version in table.columns, f"Column {c_version} not found."
        assert c_code in table.columns, f"Column {c_code} not found."
        assert table[c_version].isin(self.component_scheme_names.keys()).all(), (
            f"Only ICD version {list(self.component_scheme_names.keys())} are expected."
        )

        table = self.reformat(table, self.component_schemes(manager))

        table.loc[:, c_code] = table[c_version] + self.sep + table[c_code]

        # filter out codes that are not in the scheme.
        use_rows = table[c_code].isin(self.codes)
        removed_rows = table[~use_rows]
        removed_rows = removed_rows.assign(component_scheme=removed_rows[c_version].map(self.component_scheme_names))
        dataframe_log.info(
            f"When transforming a table to mixed code format. {len(removed_rows)} codes "
            f"were not found in the corresponding component schemes.",
            dataframe=removed_rows,
            tag="del_rows_mixed_format",
        )
        return table[use_rows].reset_index(drop=True)

    def report_lost_codes(self, dataframe: pd.DataFrame, target_name: str, mixed2target: dict[str, str]):
        lost_codes_df = dataframe[~dataframe["code"].isin(mixed2target.keys())]
        dataframe_log.info(
            f"Lost {len(lost_codes_df)} codes when generating the mapping between the Mixed  "
            f"({self.name})) and the standard ({target_name}). ",
            dataframe=lost_codes_df,
            tag=f"mixed_{self.name}_to_{target_name}_lost_codes",
        )

    def report_map_stats(
        self,
        dataframe: pd.DataFrame,
        dataframe_groupby: tuple[tuple[str, pd.DataFrame], ...],
        target_name: str,
        mixed2target: dict[str, str],
    ):
        mapped_col = f"mapped_to_{target_name}"
        subset_index = lambda name: f"{name}_subset"
        stats = pd.DataFrame(
            columns=["count", mapped_col],
            index=[subset_index(name) for name in list(zip(*dataframe_groupby))[0]] + ["total"],
        )
        stats.loc["total", "count"] = len(dataframe)
        stats.loc["total", mapped_col] = dataframe["code"].isin(mixed2target).sum()
        for name, df in dataframe_groupby:
            stats.loc[subset_index(name), "count"] = len(df)
            stats.loc[subset_index(name), mapped_col] = df["code"].isin(mixed2target).sum()

        lost_stats = stats["count"].values.reshape(-1, 1) - stats.iloc[:, 1:].rename(columns=lambda c: f"missed: {c}")
        stats = pd.concat([stats, lost_stats], axis=1)
        norm_stats = stats.rename(index=lambda i: f"%{i}") / stats["count"].values.reshape(-1, 1)
        stats = pd.concat([stats, norm_stats], axis=0)
        dataframe_log.info(
            f"Statistics of the mapping between the mixed ({self.name}) and {target_name}.",
            dataframe=stats,
            tag=f"mixed_{self.name}_to_{target_name}_stats",
        )

    def register_infer_map(self, manager: CodingSchemesManager, target_name: str) -> CodingSchemesManager:
        required_maps = tuple((component_s, target_name) for component_s in self.component_scheme_names.values())
        assert all(m in manager.map for m in required_maps), (
            f"Mapping between the mixed scheme and {target_name} is not supported due to the absence of "
            f"the map(s): {','.join(str(m) for m in required_maps if m not in manager.map)}"
        )
        dataframe = self.as_dataframe()
        dataframe_groupby = tuple((name, df) for name, df in dataframe.groupby("component_scheme"))
        mixed2target = {}
        for scheme_name, scheme_subset in dataframe_groupby:
            fmt_union_to_std = scheme_subset.set_index("code")["component_code"].to_dict()
            m = manager.map[(scheme_name, target_name)]
            mixed2target.update({mixed_c: m[comp_c] for mixed_c, comp_c in fmt_union_to_std.items() if comp_c in m})
        self.report_lost_codes(dataframe, target_name, mixed2target)
        self.report_map_stats(dataframe, dataframe_groupby, target_name, mixed2target)
        return manager.add_map(CodeMap(source_name=self.name, target_name=target_name, data=FrozenDict1N(mixed2target)))

    def register_map(
        self,
        manager: CodingSchemesManager,
        target_name: str,
        mapping: pd.DataFrame,
        c_code: str,
        c_version: str,
        c_target_code: str,
        c_target_desc: str,
    ) -> CodingSchemesManager:
        """
        Register a mapping between the current Mixed ICD scheme and a target scheme.
        """
        mapping = self.reformat(mapping.astype(str), self.component_schemes(manager))
        mapping.loc[:, c_code] = (mapping[c_version] + self.sep + mapping[c_code]).tolist()
        mapping = mapping[mapping[c_code].isin(self.codes)]
        assert len(mapping) > 0, "No mapping between the Mixed ICD scheme and the target scheme was found."
        target_codes = tuple(sorted(mapping[c_target_code].drop_duplicates().tolist()))
        target_desc = FrozenDict11(mapping.set_index(c_target_code)[c_target_desc].to_dict())
        manager = manager.add_scheme(CodingScheme(name=target_name, codes=target_codes, desc=target_desc))

        mapping = mapping[[c_code, c_target_code]].astype(str)
        mapping = mapping[mapping[c_code].isin(self.codes) & mapping[c_target_code].isin(target_codes)]
        mapping = FrozenDict1N(mapping.groupby(c_code)[c_target_code].apply(set).to_dict())
        self.report_lost_codes(self.as_dataframe(), target_name, mapping.data)
        return manager.add_map(CodeMap(source_name=self.name, target_name=target_name, data=mapping))

    def as_dataframe(self, codes: tuple[str, ...] = None) -> pd.DataFrame:
        if codes is None:
            codes = self.codes
        columns = ["code", "desc", "code_index", "component_version", "component_code"]
        table = pd.DataFrame([(c, self.desc[c], self.index[c], *c.split(self.sep)) for c in codes], columns=columns)
        table.loc[:, "component_scheme"] = list(
            map(lambda v: self.component_scheme_names[v], table["component_version"])
        )
        return table
